import numpy as np
import scipy.stats
import pytest
from astropy.io.fits import getdata


from skypy.galaxy.spectrum import dirichlet_coefficients, kcorrect_spectra


def test_sampling_coefficients():
    alpha0 = np.array([2.079, 3.524, 1.917, 1.992, 2.536])
    alpha1 = np.array([2.265, 3.862, 1.921, 1.685, 2.480])

    z1 = 1.
    redshift = np.full(1000, 2.0, dtype=float)
    redshift_reshape = np.atleast_1d(redshift)[:, np.newaxis]
    alpha = np.power(alpha0, 1. - redshift_reshape / z1) * \
        np.power(alpha1, redshift_reshape / z1)

    a0 = alpha.sum(axis=1)

    # Check the output shape if redshift is an array
    coefficients = dirichlet_coefficients(redshift, alpha0, alpha1, z1)
    assert coefficients.shape == (len(redshift), len(alpha0)), \
        'Shape of coefficients array is not (len(redshift), len(alpha0)) '

    # the marginalised distributions are beta distributions with a = alpha_i
    # and b = a0-alpha_i
    for a, c in zip(alpha.T, coefficients.T):
        d, p = scipy.stats.kstest(c, 'beta', args=(a, a0 - a))
        assert p >= 0.01, \
            'Not all marginal distributions follow a beta distribution.'

    # Test output shape if redshift is a scalar
    redshift = 2.0
    coefficients = dirichlet_coefficients(redshift, alpha0, alpha1)
    assert coefficients.shape == (len(alpha0),), \
        'Shape of coefficients array is not (len(alpha0),) ' \
        'if redshift array is float.'

    # Test raising ValueError of alpha1 and alpha0 have different size
    alpha0 = np.array([1, 2, 3])
    alpha1 = np.array([4, 5])
    redshift = np.linspace(0, 2, 10)
    with pytest.raises(ValueError):
        dirichlet_coefficients(redshift, alpha0, alpha1)

    # Test that ValueError is risen if alpha0 or alpha1 is a scalar.
    scalar_alpha = 1.
    with pytest.raises(ValueError,
                       match="alpha0 and alpha1 must be array_like."):
        dirichlet_coefficients(redshift, scalar_alpha, alpha1)
    with pytest.raises(ValueError,
                       match="alpha0 and alpha1 must be array_like."):
        dirichlet_coefficients(redshift, alpha0, scalar_alpha)


def test_kcorrect_spectra():
    # Download template data
    kcorrect_templates_url = "https://github.com/blanton144/kcorrect/raw/" \
                             "master/data/templates/k_nmf_derived.default.fits"
    lam = getdata(kcorrect_templates_url, 11)
    templates = getdata(kcorrect_templates_url, 1)

    # Test that the shape of the returned flux density corresponds to (nz, nl)
    coefficients = np.array([[0.2, 0.2, 0.2, 0.2, 0.2],
                            [0, 0.1, 0.2, 0.3, 0.4]])
    z = np.array([0.5, 1])
    mass = np.array([5 * 10 ** 10, 7 * 10 ** 9])
    lam_o, sed = kcorrect_spectra(z, mass, coefficients)

    assert sed.shape == (len(z), len(lam))

    # Test that for redshift=0, mass=1 and coefficients=[1,0,0,0,0]
    # the returned wavelengths and spectra match the template data

    coefficients = np.array([1, 0, 0, 0, 0])

    z = np.array([0])
    mass = np.array([1])

    lam_o, sed = kcorrect_spectra(z, mass, coefficients)

    assert np.allclose(lam_o, lam)
    assert np.allclose(sed, templates[0])



"""
This module contains methods that model the properties of dark matter halo
populations.
"""

from . import abundance_matching  # noqa F401,F403
from . import mass  # noqa F401,F403
from . import quenching  # noqa F401,F403



import numpy as np
from astropy import units
from scipy import stats
import pytest

from astropy.cosmology import FlatLambdaCDM
from skypy.galaxy import size


def test_angular_size():
    """ Test a FlatLambdaCDM cosmology with omega_matter = 1"""
    cosmology = FlatLambdaCDM(Om0=1.0, H0=70.0)

    # Test that a scalar input gives a scalar output
    scalar_radius = 1.0 * units.kpc
    scalar_redshift = 1.0
    angular_size = size.angular_size(scalar_radius, scalar_redshift, cosmology)

    assert np.isscalar(angular_size.value)

    # Test that the output has the correct units
    assert angular_size.unit.is_equivalent(units.rad)

    # If the input have bad units, a UnitConversionError is raised
    radius_without_units = 1.0

    with pytest.raises(units.UnitTypeError):
        size.angular_size(radius_without_units, scalar_redshift, cosmology)


def test_late_type_lognormal():
    """ Test lognormal distribution of late-type galaxy sizes"""
    # Test that a scalar input gives a scalar output
    magnitude_scalar = -20.0
    alpha, beta, gamma, M0 = 0.21, 0.53, -1.31, -20.52
    sigma1, sigma2 = 0.48, 0.25
    size_scalar = size.late_type_lognormal(magnitude_scalar, alpha, beta,
                                           gamma, M0, sigma1, sigma2)

    assert np.isscalar(size_scalar.value)

    # Test that the output has the correct units
    assert size_scalar.unit.is_equivalent(units.kpc)

    # Test that an array input gives an array output, with the same shape
    magnitude_array = np.array([-20.0, -21.0])
    size_array = size.late_type_lognormal(magnitude_array, alpha, beta, gamma,
                                          M0, sigma1, sigma2)

    assert np.shape(size_array.value) == np.shape(magnitude_array)

    # Test that size not None gives an array output, with the correct shape
    size_sample = size.late_type_lognormal(magnitude_scalar, alpha, beta,
                                           gamma, M0, sigma1, sigma2,
                                           size=1000)

    assert np.shape(size_sample.value) == (1000,)

    # Test the distribution of galaxy sizes follows a lognormal distribution
    mean = -0.4 * alpha * magnitude_scalar + (beta - alpha) *\
        np.log10(1 + np.power(10, -0.4 * (magnitude_scalar - M0)))\
        + gamma
    sigma = sigma2 + (sigma1 - sigma2) /\
        (1.0 + np.power(10, -0.8 * (magnitude_scalar - M0)))

    arguments = (sigma, 0, np.power(10, mean))
    d, p = stats.kstest(size_sample, 'lognorm', args=arguments)

    assert p > 0.01


def test_early_type_lognormal():
    """ Test lognormal distribution of late-type galaxy sizes"""
    # Test that a scalar input gives a scalar output
    magnitude_scalar = -20.0
    a, b, M0 = 0.6, -4.63, -20.52
    sigma1, sigma2 = 0.48, 0.25
    size_scalar = size.early_type_lognormal(magnitude_scalar, a, b, M0,
                                            sigma1, sigma2)

    assert np.isscalar(size_scalar.value)

    # Test that the output has the correct units
    assert size_scalar.unit.is_equivalent(units.kpc)

    # Test that an array input gives an array output, with the same shape
    magnitude_array = np.array([-20.0, -21.0])
    size_array = size.early_type_lognormal(magnitude_array, a, b, M0,
                                           sigma1, sigma2)

    assert np.shape(size_array.value) == np.shape(magnitude_array)

    # Test that size not None gives an array output, with the correct shape
    size_sample = size.early_type_lognormal(magnitude_scalar, a, b, M0,
                                            sigma1, sigma2, size=1000)

    assert np.shape(size_sample.value) == (1000,)

    # Test the distribution of galaxy sizes follows a lognormal distribution
    mean = -0.4 * a * magnitude_scalar + b
    sigma = sigma2 + (sigma1 - sigma2) /\
                     (1.0 + np.power(10, -0.8 * (magnitude_scalar - M0)))

    arguments = (sigma, 0, np.power(10, mean))
    d, p = stats.kstest(size_sample, 'lognorm', args=arguments)

    assert p > 0.01


def test_linear_lognormal():
    """ Test lognormal distribution of galaxy sizes"""
    # Test that a scalar input gives a scalar output
    magnitude_scalar = -20.0
    a_mu, b_mu, sigma = -0.24, -4.63, 0.4
    size_scalar = size.linear_lognormal(magnitude_scalar, a_mu, b_mu, sigma)

    assert np.isscalar(size_scalar.value)

    # Test that the output has the correct units
    assert size_scalar.unit.is_equivalent(units.kpc)

    # Test that an array input gives an array output, with the same shape
    magnitude_array = np.array([-20.0, -21.0])
    size_array = size.linear_lognormal(magnitude_array, a_mu, b_mu, sigma)

    assert np.shape(size_array.value) == np.shape(magnitude_array)

    # Test that size not None gives an array output, with the correct shape
    size_sample = size.linear_lognormal(magnitude_scalar, a_mu, b_mu, sigma,
                                        size=1000)

    assert np.shape(size_sample.value) == (1000,)

    # Test the distribution of galaxy sizes follows a lognormal distribution
    mean = a_mu * magnitude_scalar + b_mu
    arguments = (sigma, 0, np.power(10, mean))
    d, p = stats.kstest(size_sample, 'lognorm', args=arguments)

    assert p > 0.01



"""Models of galaxy stellar mass.
"""

import numpy as np

from skypy.utils.random import schechter


__all__ = [
    'schechter_smf',
]


def schechter_smf(alpha, m_star, x_min, x_max, resolution=100, size=None):
    r""" Stellar masses following the Schechter mass function [1]_.

    Parameters
    ----------
    alpha : float
        The alpha parameter in the Schechter stellar mass function.
    m_star : (nm,) array-like
        Characteristic stellar mass M_*.
    size: int, optional
         Output shape of stellar mass samples. If size is None and m_star
         is a scalar, a single sample is returned. If size is None and
         m_star is an array, an array of samples is returned with the same
         shape as m_star.
    x_min, x_max : float
        Lower and upper bounds for the random variable x in units of M_*.
    resolution : int, optional
        Resolution of the inverse transform sampling spline. Default is 100.

    Returns
    -------
    stellar mass : (nm,) array_like
        Drawn stellar masses from the Schechter stellar mass function in units
        of the solar mass.

    Notes
    -----
    The stellar mass probability distribution (pdf) follows a Schechter
    profile of the form

    .. math::

        \Phi(M) = \frac{1}{M_*} \left(\frac{M}{M_*}\right)^\alpha
            \exp\left(-\frac{M}{M_*}\right) \;.

    From this pdf one can sample the stellar masses.

    Examples
    --------
    >>> from skypy.galaxy import stellar_mass

    Sample 100 stellar masses values at redshift z = 1.0 with alpha = -1.4,
    m_star = 10**10.67, x_min = 0.0002138 and x_max = 213.8

    >>> masses = stellar_mass.schechter_smf(-1.4, 10**10.67, 0.0002138,
    ...                               213.8, size=100)

    References
    ----------
    .. [1] Mo, H., Van den Bosch, F., & White, S. (2010). Galaxy Formation and
        Evolution. Cambridge: Cambridge University Press.
        doi:10.1017/CBO9780511807244

    """

    if size is None and np.shape(m_star):
        size = np.shape(m_star)

    x_sample = schechter(alpha, x_min, x_max, resolution=resolution, size=size)

    return m_star * x_sample



"""
``skypy`` Command Line Script
=============================

``skypy`` is a command line script that runs a pipeline of functions defined in
a config file to generate tables of objects and write them to file.

Using ``skypy`` to run one of the example pipelines and write the outputs to
fits files:

    $ skypy examples/mccl_galaxies.yml --format fits

Config Files
------------

Config files are written in yaml format. The top level should contain the
fields ``cosmology`` and/or ``tables``. ``cosmology`` should contain a
dictionary configuring a function that returns an
``astropy.cosmology.Cosmology`` object. ``tables`` should contain a set of
nested dictionaries, first giving the names of each table, then the names of
each column within each table. Each column should contain a dictionary
configuring a function that returns an array-like object.

Each step in the pipeline is configured by a dictionary specifying:

- 'function' : the name of the function
- 'module' : the name of the the module to import 'function' from
- 'args' : a list of positional arguments (by value)
- 'kwargs' : a dictionary of keyword arguments
- 'requires' : a dictionary of keyword arguments

Note that 'kwargs' specifices keyword arguments by value, wheras 'requires'
specifices the names of previous steps in the pipeline and uses their return
values as keyword arguments.
"""

import argparse
from skypy import __version__ as skypy_version
from skypy.pipeline.driver import SkyPyDriver
import sys


def yaml_tag(loader, tag, node):
    '''handler for generic YAML tags

    tags are stored as a tuple `(tag, value)`
    '''

    import yaml

    if isinstance(node, yaml.ScalarNode):
        value = loader.construct_scalar(node)
    elif isinstance(node, yaml.SequenceNode):
        value = loader.construct_sequence(node)
    elif isinstance(node, yaml.MappingNode):
        value = loader.construct_mapping(node)

    # tags without arguments have empty string value
    if value == '':
        return tag,

    return tag, value


def main(args=None):

    import yaml

    parser = argparse.ArgumentParser(description="SkyPy pipeline driver")
    parser.add_argument('--version', action='version', version=skypy_version)
    parser.add_argument('config', type=argparse.FileType('r'),
                        help='Config file name')
    parser.add_argument('-f', '--format', required=False,
                        choices=['fits', 'hdf5'], help='Table file format')
    parser.add_argument('-o', '--overwrite', action='store_true',
                        help='Whether to overwrite existing files')

    # get system args if none passed
    if args is None:
        args = sys.argv[1:]

    # register custom tags
    yaml.SafeLoader.add_multi_constructor('!', yaml_tag)

    args = parser.parse_args(args or ['--help'])
    config = yaml.safe_load(args.config)
    config = {} if config is None else config
    driver = SkyPyDriver()
    driver.execute(config, file_format=args.format, overwrite=args.overwrite)
    return(0)



"""Astronomy utility module.

This module provides methods to convert among astronomical quantities
like luminosity and magnitude.

Utility functions
=================

.. autosummary::
   :nosignatures:
   :toctree: ../api/

   luminosity_from_absolute_magnitude
   absolute_magnitude_from_luminosity

"""

import numpy as np


def luminosity_from_absolute_magnitude(absolute_magnitude, zeropoint=None):
    """ Converts absolute magnitudes into luminosities

    Parameters
    ----------
    absolute_magnitude : array_like
        Input absolute magnitudes
    zeropoint : float, optional
        Zeropoint for the conversion.

    Returns
    -------
    ndarray, or float if input is scalar
    Luminosity values.
    """

    if zeropoint is None:
        zeropoint = 0.

    return 10.**(-0.4*np.add(absolute_magnitude, zeropoint))


def absolute_magnitude_from_luminosity(luminosity, zeropoint=None):
    """ Converts luminosities into absolute magnitudes

    Parameters
    ----------
    luminosity : array_like
        Input luminosity
    zeropoint : float, optional
        Zeropoint for the conversion.

    Returns
    -------
    ndarray, or float if input is scalar
    Absolute magnitude values.
    """

    if zeropoint is None:
        zeropoint = 0.

    return -2.5*np.log10(luminosity) - zeropoint



import skypy.utils.astronomy as astro


def test_convert_abs_mag_to_lum():
    assert round(astro.luminosity_from_absolute_magnitude(-22), 1) \
           == 630957344.5


def test_convert_lum_to_abs_mag():
    assert round(astro.absolute_magnitude_from_luminosity(630957344.5), 0) \
           == -22



from astropy.utils.data import get_pkg_data_filename
from contextlib import redirect_stdout
from io import StringIO
import pytest
from skypy import __version__ as skypy_version
from skypy.pipeline.scripts import skypy


def test_skypy():

    # No arguments
    with pytest.raises(SystemExit) as e:
        skypy.main([])
    assert e.value.code == 0

    # Argparse help
    with pytest.raises(SystemExit) as e:
        skypy.main(['--help'])
    assert e.value.code == 0

    # Argparse version
    version = StringIO()
    with pytest.raises(SystemExit) as e:
        with redirect_stdout(version):
            skypy.main(['--version'])
    assert version.getvalue().strip() == skypy_version
    assert e.value.code == 0

    # Missing positional argument 'config'
    with pytest.raises(SystemExit) as e:
        skypy.main(['--format', 'fits'])
    assert e.value.code == 2

    # Invalid file format
    with pytest.raises(SystemExit) as e:
        skypy.main(['--format', 'invalid', 'config.filename'])
    assert e.value.code == 2

    # Process empty config file
    filename = get_pkg_data_filename('data/empty_config.yml')
    assert skypy.main([filename]) == 0

    # Process test config file
    filename = get_pkg_data_filename('data/test_config.yml')
    assert skypy.main([filename]) == 0



import numpy as np
from astropy.cosmology import Planck15
from astropy.units import allclose
import pytest
from skypy.power_spectrum import halofit_smith
from skypy.power_spectrum import halofit_takahashi
from skypy.power_spectrum import halofit_bird


def test_halofit():
    """Test Smith, Takahashi and Bird Halofit models with Planck15 cosmology"""

    # Data for tests
    k = np.logspace(-4.0, 0.0, 5)
    z = np.linspace(0.0, 1.0, 2)
    p = [[705.54997046, 262.14967329], [6474.60158058, 2405.66190924],
         [37161.00990355, 13807.30920991], [9657.02613688, 3588.10339832],
         [114.60445565, 42.58170486]]
    ts = [[705.49027968, 262.13980368], [6469.19607307, 2404.75754883],
          [36849.24061946, 13757.68241714], [9028.01112208, 3628.67740715],
          [596.91685425, 110.08074646]]
    tt = [[705.48895748, 262.14055831], [6469.02581579, 2404.83678008],
          [36827.71983838, 13751.52554662], [9143.97447325, 3050.69467676],
          [662.31133378, 60.66609697]]
    tb = [[705.4903004, 262.13980169], [6469.19940132, 2404.75760398],
          [36849.2113973, 13757.8052238], [9010.83583125, 3630.48463588],
          [601.45212141, 111.52139435]]
    linear_power = np.array(p)
    truth_smith = np.array(ts)
    truth_takahashi = np.array(tt)
    truth_bird = np.array(tb)

    # Non-linear power spectra from Smith, Takahashi and Bird models
    nl_power_smith = halofit_smith(k, z, linear_power, Planck15)
    nl_power_takahashi = halofit_takahashi(k, z, linear_power, Planck15)
    nl_power_bird = halofit_bird(k, z, linear_power, Planck15)

    # Test shape of outputs
    assert np.shape(nl_power_smith) == np.shape(linear_power)
    assert np.shape(nl_power_takahashi) == np.shape(linear_power)
    assert np.shape(nl_power_bird) == np.shape(linear_power)

    # Test outputs against precomputed values
    assert allclose(nl_power_smith, truth_smith)
    assert allclose(nl_power_takahashi, truth_takahashi)
    assert allclose(nl_power_bird, truth_bird)

    # Test when redshift is a scalar
    z_scalar = z[0]
    power_1d = linear_power[:, 0]
    truth_scalar_redshift = truth_smith[:, 0]
    smith_scalar_redshift = halofit_smith(k, z_scalar, power_1d, Planck15)
    assert allclose(smith_scalar_redshift, truth_scalar_redshift)
    assert np.shape(smith_scalar_redshift) == np.shape(power_1d)

    # Test for failure when wavenumber is a scalar
    k_scalar = k[0]
    power_1d = linear_power[0, :]
    with pytest.raises(TypeError):
        halofit_smith(k_scalar, z, power_1d, Planck15)

    # Test for failure when wavenumber array is the wrong size
    k_wrong_size = np.logspace(-4.0, 2.0, 7)
    with pytest.raises(ValueError):
        halofit_smith(k_wrong_size, z, linear_power, Planck15)

    # Test for failure when redshift arry is the wrong size
    z_wrong_size = np.linspace(0.0, 2.0, 3)
    with pytest.raises(TypeError):
        halofit_smith(k, z_wrong_size, linear_power, Planck15)

    # Test for failure when wavenumber is negative
    k_negative = np.copy(k)
    k_negative[0] = -1.0
    with pytest.raises(ValueError):
        halofit_smith(k_negative, z, linear_power, Planck15)

    # Test for failure when wavenumber is zero
    k_zero = np.copy(k)
    k_zero[0] = 0.0
    with pytest.raises(ValueError):
        halofit_smith(k_zero, z, linear_power, Planck15)

    # Test for failure when redshift is negative
    z_negative = np.copy(z)
    z_negative[0] = -1.0
    with pytest.raises(ValueError):
        halofit_smith(k, z_negative, linear_power, Planck15)

    # Test for failure when linear power spectrum is negative
    power_negative = np.copy(linear_power)
    power_negative[0, 0] = -1.0
    with pytest.raises(ValueError):
        halofit_smith(k, z, power_negative, Planck15)

    # Test for failure when wavenumber array is not in asscending order
    k_wrong_order = k[::-1]
    with pytest.raises(ValueError):
        halofit_smith(k_wrong_order, z, linear_power, Planck15)



from astropy.cosmology import default_cosmology
from astropy.io import fits
from astropy.table import Table
import networkx
import numpy as np
import os
import pytest
from skypy.pipeline.driver import SkyPyDriver


def test_driver():

    # Evaluate and store the default astropy cosmology.
    config = {'test_cosmology': ('astropy.cosmology.default_cosmology.get',)}

    driver = SkyPyDriver()
    driver.execute(config)
    assert driver.test_cosmology == default_cosmology.get()

    # Generate a simple two column table with a dependency. Also write the
    # table to a fits file and check it's contents.
    size = 100
    string = size*'a'
    config = {'tables': {
                'test_table': {
                  'column1': ('numpy.random.uniform', {
                      'size': size}),
                  'column2': ('numpy.random.uniform', {
                      'low': ('test_table.column1',)}),
                  'column3': ('list', [string])}}}

    driver = SkyPyDriver()
    driver.execute(config, file_format='fits')
    assert len(driver.test_table) == size
    assert np.all(driver.test_table['column1'] < driver.test_table['column2'])
    with fits.open('test_table.fits') as hdu:
        assert np.all(Table(hdu[1].data) == driver.test_table)

    # Check for failure if output files already exist and overwrite is False
    driver = SkyPyDriver()
    with pytest.raises(OSError):
        driver.execute(config, file_format='fits', overwrite=False)

    # Check that the existing output files are modified if overwrite is True
    new_size = 2 * size
    new_string = new_size*'a'
    config['tables']['test_table']['column1'][1]['size'] = new_size
    config['tables']['test_table']['column3'][1][0] = new_string
    driver = SkyPyDriver()
    driver.execute(config, file_format='fits', overwrite=True)
    with fits.open('test_table.fits') as hdu:
        assert len(hdu[1].data) == new_size

    # Check for failure if 'column1' calls a nonexistant module
    config['tables']['test_table']['column1'] = ('nomodule.function',)
    with pytest.raises(ModuleNotFoundError):
        SkyPyDriver().execute(config)

    # Check for failure if 'column1' calls a nonexistant function
    config['tables']['test_table']['column1'] = ('builtins.nofunction',)
    with pytest.raises(AttributeError):
        SkyPyDriver().execute(config)

    # Check for failure if 'column1' requires itself creating a cyclic
    # dependency graph
    config['tables']['test_table']['column1'] = ('list', ('test_table.column1',))
    with pytest.raises(networkx.NetworkXUnfeasible):
        SkyPyDriver().execute(config)

    # Check for failure if 'column1' and 'column2' both require each other
    # creating a cyclic dependency graph
    config['tables']['test_table']['column1'] = ('list', ('test_table.column2',))
    with pytest.raises(networkx.NetworkXUnfeasible):
        SkyPyDriver().execute(config)

    # Check for failure if 'column1' is removed from the config so that the
    # requirements for 'column2' are not satisfied.
    del config['tables']['test_table']['column1']
    with pytest.raises(KeyError):
        SkyPyDriver().execute(config)

    # Check variables intialised by value
    config = {'test_int': 1,
              'test_float': 1.0,
              'test_string': 'hello world',
              'test_list': [0, 'one', 2.],
              'test_dict': {'a': 'b'}}
    driver = SkyPyDriver()
    driver.execute(config)
    assert isinstance(driver.test_int, int)
    assert isinstance(driver.test_float, float)
    assert isinstance(driver.test_string, str)
    assert isinstance(driver.test_list, list)
    assert isinstance(driver.test_dict, dict)
    assert driver.test_int == 1
    assert driver.test_float == 1.0
    assert driver.test_string == 'hello world'
    assert driver.test_list == [0, 'one', 2.]
    assert driver.test_dict == {'a': 'b'}

    # Check variables intialised by function
    config = {'test_func': ('list', 'hello world'),
              'len_of_test_func': ('len', ('test_func',)),
              'nested_references': ('sum', [
                [('test_func',), [' '], ('test_func',)], []])}
    driver = SkyPyDriver()
    driver.execute(config)
    assert driver.test_func == list('hello world')
    assert driver.len_of_test_func == len('hello world')
    assert driver.nested_references == list('hello world hello world')


def teardown_module(module):

    # Remove fits file generated in test_driver
    os.remove('test_table.fits')



"""Random utility module.

This module provides methods to draw from random distributions.


Utility Functions
=================

.. autosummary::
   :nosignatures:
   :toctree: ../api/

   schechter

"""

import numpy as np
import skypy.utils.special as special


def schechter(alpha, x_min, x_max, resolution=100, size=None):
    """Sample from the Schechter function.

    Parameters
    ----------
    alpha : float or int
        The alpha parameter in the Schechter function in [1]_.
    x_min, x_max : array_like
        Lower and upper bounds for the random variable x.
    resolution : int
        Resolution of the inverse transform sampling spline. Default is 100.
    size: int, optional
        Output shape of samples. Default is None.

    Returns
    -------
    x_sample : array_like
        Samples drawn from the Schechter function.

    Examples
    --------
    >>> import skypy.utils.random as random
    >>> alpha = -1.3
    >>> sample = random.schechter(alpha, x_min=1e-10, x_max=1e2,
    ...                           resolution=100, size=1000)


    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Luminosity_function_(astronomy)
    """

    x = np.logspace(np.log10(np.min(x_min)), np.log10(np.max(x_max)),
                    resolution)
    cdf = _schechter_cdf(x, np.min(x_min), np.max(x_max), alpha)
    t_lower = np.interp(x_min, x, cdf)
    t_upper = np.interp(x_max, x, cdf)
    u = np.random.uniform(t_lower, t_upper, size=size)
    x_sample = np.interp(u, cdf, x)

    return x_sample


def _schechter_cdf(x, x_min, x_max, alpha):
    a = special.gammaincc(alpha + 1, x_min)
    b = special.gammaincc(alpha + 1, x)
    c = special.gammaincc(alpha + 1, x_min)
    d = special.gammaincc(alpha + 1, x_max)

    return (a - b) / (c - d)



import numpy as np
import scipy.stats
import scipy.integrate
import scipy.special
import pytest

import skypy.galaxy.luminosity as lum
import skypy.utils.special as special
import skypy.utils.astronomy as astro


def test_calculate_luminosity_star():
    luminosity_star = lum._calculate_luminosity_star(1, 2.5, 5)
    assert luminosity_star == 0.001

    redshift = np.array([0.1, 0.6, 1])
    luminosity_star = lum._calculate_luminosity_star(redshift, 2.5, 5)
    result = np.array([0.00794328, 0.00251189, 0.001])
    np.testing.assert_allclose(luminosity_star, result, rtol=1e-05)


def test_herbel_luminosities():
    # Test that error is returned if redshift input is an array but size !=
    # None and size != redshift,size
    with pytest.raises(ValueError):
        lum.herbel_luminosities(np.array([1, 2]), -1.3, -0.9408582,
                                -20.40492365, size=3)

    # Test that sampling corresponds to sampling from the right pdf.
    # For this, we sample an array of luminosities for redshift z = 1.0 and we
    # compare it to the corresponding cdf.
    # The B-band zeropoint for the model is -4.73.

    def calc_pdf(luminosity, z, alpha, a_m, b_m, absolute_magnitude_max=-16.0,
                 absolute_magnitude_min=-28.0):
        luminosity_min = astro.luminosity_from_absolute_magnitude(
            absolute_magnitude_max, -4.73)
        luminosity_max = astro.luminosity_from_absolute_magnitude(
            absolute_magnitude_min, -4.73)
        lum_star = lum._calculate_luminosity_star(z, a_m, b_m, -4.73)
        c = np.fabs(special.gammaincc(alpha+1, luminosity_min/lum_star))
        d = np.fabs(special.gammaincc(alpha+1, luminosity_max/lum_star))
        return 1./lum_star*np.power(luminosity/lum_star, alpha)*np.exp(
                -luminosity/lum_star-scipy.special.gammaln(alpha+1))/(c - d)

    def calc_cdf(L):
        a_m = -0.9408582
        b_m = -20.40492365
        alpha = -1.3
        pdf = calc_pdf(L, 1.0, alpha, a_m, b_m)
        cdf = scipy.integrate.cumtrapz(pdf, L, initial=0)
        cdf = cdf / cdf[-1]
        return cdf

    sample = lum.herbel_luminosities(1.0, -1.3, -0.9408582, -20.40492365,
                                     size=1000)
    p_value = scipy.stats.kstest(sample, calc_cdf)[1]
    assert p_value >= 0.01


def test_exponential_distribution():
    # When alpha=0, L*=1 and x_min~0 we get a truncated exponential
    # the b_m value offsets the B-band zeropoint of -4.73
    q_max = 1e2
    sample = lum.herbel_luminosities(0, 0, 0, 4.73, size=1000,
                                     x_min=1e-10, x_max=q_max,
                                     resolution=1000)
    d, p_value = scipy.stats.kstest(sample, 'truncexpon', args=(q_max,))
    assert p_value >= 0.01



"""Special functions.

This module computes useful special functions.

.. autosummary::
   :nosignatures:
   :toctree: ../api/

   gammaincc

"""

import numpy as np
import scipy.special as sc


def _gammaincc(a, x):
    '''gammaincc for positive or negative indices and scalar inputs'''
    if x < 0:
        raise ValueError('negative x in gammaincc')
    if x == 0:
        if a > 0:
            return 1
        s = sc.gammasgn(a)
        if s == 0:
            return 0
        return s*np.inf
    if np.isinf(x):
        return 0
    if a < 0:
        n = np.floor(a)
    else:
        n = 0
    g = sc.gammaincc(a-n, x) if a != n else 0
    if n < 0:
        f = np.exp(sc.xlogy(a-n, x)-x-sc.gammaln(a-n+1))
        while n < 0:
            f *= (a-n)/x
            g -= f
            n += 1
    return g


def gammaincc(a, x):
    r'''Regularized upper incomplete gamma function.

    This implementation of `gammaincc` allows :math:`a` real and :math:`x`
    nonnegative.

    Parameters
    ----------
    a : array_like
        Real parameter.

    x : array_like
        Nonnegative argument.

    Returns
    -------
    scalar or ndarray
        Values of the upper incomplete gamma function.

    Notes
    -----
    The function value is computed via a recurrence from the value of
    `scipy.special.gammaincc` for arguments :math:`a-n, x` where :math:`n` is
    the smallest integer such that :math:`a-n \ge 0`.

    See also
    --------
    scipy.special.gammaincc : Computes the start of the recurrence.

    Examples
    --------
    This implementation of `gammaincc` supports positive and negative values
    of `a`.

    >>> from skypy.utils.special import gammaincc
    >>> gammaincc([-1.5, -0.5, 0.5, 1.5], 1.2)
    array([ 0.03084582, -0.03378949,  0.12133525,  0.49363462])

    '''
    if np.broadcast(a, x).ndim == 0:
        return _gammaincc(a, x)
    a, x = np.broadcast_arrays(a, x)
    if np.any(x < 0):
        raise ValueError('negative x in gammaincc')

    # nonpositive a need special treatment
    i = a <= 0

    # find integer n such that a + n >= 0
    n = np.where(i, np.floor(a), 0)

    # compute gammaincc for a-n and x as usual
    g = np.where(a == n, 0, sc.gammaincc(a-n, x))

    # deal with nonpositive a
    # the number n keeps track of iterations still to do
    if np.any(i):
        # all x = inf are done
        n[i & np.isinf(x)] = 0

        # do x = 0 for nonpositive a; depends on Gamma(a)
        i = i & (x == 0)
        s = sc.gammasgn(a[i])
        g[i] = np.where(s == 0, 0, np.copysign(np.inf, s))
        n[i] = 0

        # these are still to do
        i = n < 0

        # recurrence
        f = np.empty_like(g)
        f[i] = np.exp(sc.xlogy(a[i]-n[i], x[i])-x[i]-sc.gammaln(a[i]-n[i]+1))
        while np.any(i):
            f[i] *= (a[i]-n[i])/x[i]
            g[i] -= f[i]
            n[i] += 1
            i[i] = n[i] < 0
    return g




r"""Binary merger rate module.
This module provides functions to calculate compact binary merger rates
for individual galaxies.

"""

from astropy import constants, units


__all__ = [
    'b_band_merger_rate',
]


abadie_table_III = {
    'NS-NS': {
        'low': 0.6,
        'realistic': 60,
        'high': 600,
        'max': 2000},
    'NS-BH': {
        'low': 0.03,
        'realistic': 2,
        'high': 60},
    'BH-BH': {
        'low': 0.006,
        'realistic': 0.2,
        'high': 20}
}


def b_band_merger_rate(luminosity,
                       population='NS-NS',
                       optimism='low'):

    r"""Model of Abadie et al (2010), Table III

    Compact binary merger rates as a linear function of a galaxies
    B-band luminosity.

    Parameters
    ----------
    luminosity : (ngal,) array-like
        The B-band luminosity of the galaxies to generate merger
        rates for, in units of solar luminosity.
    population : {'NS-NS', 'NS-BH', 'BH-BH'}
        Compact binary population to get rate for.
        'NS-NS' is neutron star - neutron star
        'NS-BH' is neutron star - black hole
        'BH-BH' is black hole - black hole
    optimism : {'low', 'realistic', 'high'}
        Optimism of predicted merger rates.
        For 'NS-NS' there is an extra option 'max'.

    Returns
    -------
    merger_rate : array_like
        Merger rates for the galaxies in units of year^-1

    Notes
    -----

    References
    ----------
    .. Abadie et al. 2010, Classical and Quantum Gravity,
        Volume 27, Issue 17, article id. 173001 (2010)
        https://arxiv.org/abs/1003.2480

    Examples
    --------
    >>> from skypy.galaxy.luminosity import herbel_luminosities
    >>> from skypy.gravitational_wave import b_band_merger_rate

    Sample 100 luminosity values at redshift z = 1.0 with
    a_m = -0.9408582, b_m = -20.40492365, alpha = -1.3.

    >>> luminosities = herbel_luminosities(1.0, -1.3, -0.9408582,
    ...                                    -20.40492365, size=100)

    Generate merger rates for these luminosities.

    >>> rates = b_band_merger_rate(luminosities,
    ...                            population='NS-NS',
    ...                            optimism='low')

    """

    # Convert luminosity to units of L_10 defined in Abadie et. al. 2010
    L_10 = luminosity * constants.L_sun.to_value('erg/s') / (1e10 * 2.16e33)
    return abadie_table_III[population][optimism] * L_10 / units.year



import numpy as np

from skypy.galaxy.redshift import smail, herbel_redshift, herbel_pdf
from unittest.mock import patch
from scipy.stats import kstest, norm
import scipy.integrate


def test_smail():
    # sample a single redshift
    rvs = smail(1.3, 2.0, 1.5)
    assert np.isscalar(rvs)

    # sample 10 reshifts
    rvs = smail(1.3, 2.0, 1.5, size=10)
    assert rvs.shape == (10,)

    # sample with implicit sizes
    zm, a, b = np.ones(5), np.ones((7, 5)), np.ones((13, 7, 5))
    rvs = smail(z_median=zm, alpha=a, beta=b)
    assert rvs.shape == np.broadcast(zm, a, b).shape

    # check sampling, for alpha=0, beta=1, the distribution is exponential
    D, p = kstest(smail(0.69315, 1e-100, 1., size=1000), 'expon')
    assert p > 0.01, 'D = {}, p = {}'.format(D, p)

    # check sampling, for beta=1, the distribution matches a gamma distribution
    D, p = kstest(smail(2.674, 2, 1, size=1000), 'gamma', args=(3,))
    assert p > 0.01, 'D = {}, p = {}'.format(D, p)

    # check sampling, the distribution is a generalised gamma distribution
    D, p = kstest(smail(0.832555, 1, 2, size=1000), 'gengamma', args=(1, 2))
    assert p > 0.01, 'D = {}, p = {}'.format(D, p)


def test_herbel_pdf():
    from astropy.cosmology import FlatLambdaCDM
    cosmology = FlatLambdaCDM(H0=70, Om0=0.3, Tcmb0=2.725)
    pdf = herbel_pdf(np.array([0.01, 0.5, 1, 2]),
                     -0.5, -0.70596888,
                     0.0035097, -0.70798041,
                     -20.37196157, cosmology, -16.0)
    result = np.array(
        [4.09063927e+04, 4.45083420e+07, 7.26629445e+07, 5.40766813e+07])
    np.testing.assert_allclose(pdf, result)


# Test whether principle of the interpolation works. Let PDF return the PDF
# of a Gaussian and sample from the corresponding CDF. Then compare the
# first three moment of the returned sample with the Gaussian one.
@patch('skypy.galaxy.redshift.herbel_pdf')
def test_herbel_redshift_gauss(herbel_pdf):
    from astropy.cosmology import FlatLambdaCDM
    cosmology = FlatLambdaCDM(H0=70, Om0=0.3, Tcmb0=2.725)
    resolution = 100
    x = np.linspace(-5, 5, resolution)
    herbel_pdf.side_effect = [norm.pdf(x)]
    sample = herbel_redshift(alpha=-1.3, a_phi=-0.10268436,
                             a_m=-0.9408582, b_phi=0.00370253,
                             b_m=-20.40492365, cosmology=cosmology,
                             low=-5., high=5.0, size=1000000,
                             resolution=resolution)
    p_value = kstest(sample, 'norm')[1]
    assert p_value >= 0.01


# Test that the sampling follows the pdf of Schechter function.
def test_herbel_redshift_sampling():
    from astropy.cosmology import FlatLambdaCDM
    cosmology = FlatLambdaCDM(H0=70, Om0=0.3, Tcmb0=2.725)
    sample = herbel_redshift(alpha=-1.3, a_phi=-0.10268436,
                             a_m=-0.9408582, b_phi=0.00370253,
                             b_m=-20.40492365, cosmology=cosmology,
                             size=1000)

    def calc_cdf(z):
        pdf = herbel_pdf(z, alpha=-1.3,
                         a_phi=-0.10268436,
                         a_m=-0.9408582,
                         b_phi=0.00370253,
                         b_m=-20.40492365,
                         cosmology=cosmology,
                         absolute_magnitude_max=-16)
        cdf = scipy.integrate.cumtrapz(pdf, z, initial=0)
        cdf = cdf / cdf[-1]
        return cdf

    p_value = kstest(sample, calc_cdf)[1]
    assert p_value >= 0.01



"""Galaxy quenching.

This module implements models for environment and mass
quenching by dark matter halos.

Models
======
.. autosummary::
   :nosignatures:
   :toctree: ../api/

   environment_quenched
   mass_quenched
"""

import numpy as np
from scipy import special

__all__ = [
    'environment_quenched',
    'mass_quenched',
    ]


def environment_quenched(nh, probability):
    r'''Environment quenching function.
    This function implements the model proposed by A.Amara where the
    probability of a subhalo being quenched is a fixed
    probability. The model is inspired on [1]_ and [2]_.

    Parameters
    ----------
    nh: integer
        Number of subhalos.
    probability: float
        Quenching probability.

    Returns
    -------
    quenched: (nh,) array_like,  boolean
        Boolean array indicating which subhalo's host galaxies are
        (satellite) environment-quenched.

    Examples
    ---------

    This example shows how many subhalos are environtment quenched (True)
    and how many survive (False) from a list of 1000 halos:

    >>> import numpy as np
    >>> from skypy.halo.quenching import environment_quenched
    >>> from collections import Counter
    >>> quenched = environment_quenched(1000, 0.5)
    >>> Counter(quenched)
    Counter({...})

    References
    ----------
    .. [1] Peng et al. 2010, doi 10.1088/0004-637X/721/1/193.
    .. [2] Birrer et al. 2014, arXiv 1401.3162.

    '''

    return np.random.uniform(size=nh) < probability


def mass_quenched(halo_mass, offset, width):
    r'''Mass quenching function.
    This function implements the model proposed by A.Amara where the
    probability of a halo being quenched is related to the error function
    of the logarithm of the halo's mass standardised by an offset and width
    parameter.  The model is inspired on [1]_ and [2]_.

    Parameters
    ----------
    halo_mass: (nh,) array_like
        Array of halo masses in units of solar mass, :math:`M_{sun}`.
    offset: float
        Halo mass in :math:`M_{sun}` at which quenching probability is 50%.
    width: float
        Width of the error function.

    Returns
    -------
    quenched: (nh,) array_like, boolean
        Boolean array indicating which halo's host galaxies are mass-quenched.

    Examples
    ---------

    This example shows how many halos are mass quenched (True)
    and how many survive (False) from a list of 1000 halos:

    >>> import numpy as np
    >>> from astropy import units
    >>> from skypy.halo.quenching import mass_quenched
    >>> from collections import Counter
    >>> offset, width = 1.0e12, 0.5
    >>> halo_mass = np.random.lognormal(mean=np.log(offset), sigma=width,
    ...                                 size=1000)
    >>> quenched = mass_quenched(halo_mass, offset, width)
    >>> Counter(quenched)
    Counter({...})

    References
    ----------
    .. [1] Peng et al. 2010, doi 10.1088/0004-637X/721/1/193.
    .. [2] Birrer et al. 2014, arXiv 1401.3162.

    '''

    standardised_mass = np.log10(halo_mass / offset) / width
    probability = 0.5 * (1.0 + special.erf(standardised_mass/np.sqrt(2)))
    nh = len(halo_mass)
    return np.random.uniform(size=nh) < probability



"""
This module contains methods that model the properties of gravitational wave
populations.

Merger Rates
================

.. autosummary::
   :nosignatures:
   :toctree: ../api/

   b_band_merger_rate

"""

from .merger_rate import *  # noqa F401,F403



r"""Galaxy spectrum module.

"""

import numpy as np
from astropy.io.fits import getdata


__all__ = [
    'dirichlet_coefficients',
    'kcorrect_spectra'
]


def dirichlet_coefficients(redshift, alpha0, alpha1, z1=1.):
    r"""Dirichlet-distributed SED coefficients.

    Spectral coefficients to calculate the rest-frame spectral energy
        distribution of a galaxy following the Herbel et al. model in [1]_.

    Parameters
    ----------
    redshift : (nz,) array_like
        The redshift values of the galaxies for which the coefficients want to
        be sampled.
    alpha0, alpha1 : (nc,) array_like
        Factors parameterising the Dirichlet distribution according to Equation
        (3.9) in [1]_.
    z1 : float or scalar, optional
       Reference redshift at which alpha = alpha1. The default value is 1.0.

    Returns
    -------
    coefficients : (nz, nc) ndarray
        The spectral coefficients of the galaxies. The shape is (n, nc) with nz
        the number of redshifts and nc the number of coefficients.

    Notes
    -----
    One example is the rest-frame spectral energy distribution of galaxies
    which can be written as a linear combination of the five kcorrect [2]_
    template spectra :math:`f_i`
    (see [1]_)

    .. math::

        f(\lambda) = \sum_{i=1}^5 c_i f_i(\lambda) \;,

    where the coefficients :math:`c_i` were shown to follow a Dirichlet
    distribution of order 5. The five parameters describing the Dirichlet
    distribution are given by

    .. math::

        \alpha_i(z) = (\alpha_{i,0})^{1-z/z_1} \cdot (\alpha_{i,1})^{z/z_1} \;.

    Here, :math:`\alpha_{i,0}` describes the galaxy population at redshift
    :math:`z=0` and :math:`\alpha_{i,1}` the population at :math:`z=z_1 > 0`.
    These parameters depend on the galaxy type and we chose :math:`z_1=1`.

    Beside this example, this code works for a general number of templates.

    References
    ----------
    .. [1] Herbel J., Kacprzak T., Amara A. et al., 2017, Journal of Cosmology
           and Astroparticle Physics, Issue 08, article id. 035 (2017)
    .. [2] Blanton M. R., Roweis S., 2007, The Astronomical Journal,
           Volume 133, Page 734

    Examples
    --------
    >>> from skypy.galaxy.spectrum import dirichlet_coefficients
    >>> import numpy as np

    Sample the coefficients according to [1]_ for n blue galaxies with
    redshifts between 0 and 1.

    >>> n = 100000
    >>> alpha0 = np.array([2.079, 3.524, 1.917, 1.992, 2.536])
    >>> alpha1 = np.array([2.265, 3.862, 1.921, 1.685, 2.480])
    >>> redshift = np.linspace(0,2, n)
    >>> coefficients = dirichlet_coefficients(redshift, alpha0, alpha1)

    """
    if np.isscalar(alpha0) or np.isscalar(alpha1):
        raise ValueError("alpha0 and alpha1 must be array_like.")
    return_shape = (*np.shape(redshift), *np.shape(alpha0))
    redshift = np.atleast_1d(redshift)[:, np.newaxis]

    alpha = np.power(alpha0, 1. - redshift / z1) * \
        np.power(alpha1, redshift / z1)

    # To sample Dirichlet distributed variables of order k we first sample k
    # Gamma distributed variables y_i with the parameters alpha_i.
    # Normalising all y_i by the sum of all k y_i gives us the k Dirichlet
    # distributed variables.
    y = np.random.gamma(alpha)
    sum_y = y.sum(axis=1)
    coefficients = np.divide(y.T, sum_y.T).T
    return coefficients.reshape(return_shape)


def kcorrect_spectra(redshift, stellar_mass, coefficients):
    r"""Flux densities of galaxies.

    The flux density as a sum of the 5 kcorrect templates.

    Parameters
    ----------
    redshift : (nz,) array-like
        The redshift values of the galaxies.
    stellar_mass : (nz, ) array-like
        The stellar masses of the galaxies.
    coefficients: (nz, 5) array-like
        Coefficients to be multiplied with the kcorrect templates.


    Returns
    -------
    wavelength_observe : (nl, ) array_like
        Wavelengths corresponding to the flux density. Given in units of
        Angstrom
    sed: (nz, nl) array-like
        Flux density of the galaxies in units of erg/s/cm^2/Angstrom as it
        would be observed at a distance of 10 pc.

    Notes
    -----
    The rest-frame flux-density can be calculated as a sum of the five kcorrect
    templates [1]_

    .. math::
        f_e(\lambda) = \sum_i c_i t_i(\lambda) \;,

    with kcorrect templates :math:`t_i(\lambda)` and coefficients :math:`c_i`.

    The kcorrect templates are given in units of
    erg/s/cm^2/Angstrom per solar mass and as it would be observed in a
    distance of 10pc. To obtain the correct flux density if the object would be
    at 10 pc distance we have to adjust the coefficients by the stellar mass
    :math:`M` of the galaxy

    .. math::
         \tilde{c_i} = c_i \cdot M \;.

    Thus, the flux density is given by

    .. math::
        f_e(\lambda) = \sum_i \tilde{c_i} t_i(\lambda) \;.

    To get the flux density in observed frame we have to redshift it

    .. math::
        f_o(\lambda_o) = \frac{f_e(\lambda)}{1+z} \;

    where

    .. math::
        \lambda_o = (1+z) \lambda \;.

    References
    ----------
    .. [1] Blanton M., Roweis S., 2006, The Astronomical Journal, Issue 2,
        Volume 133, Pages 734 - 754

    Examples
    --------
    >>> from skypy.galaxy.spectrum import kcorrect_spectra
    >>> from astropy.cosmology import FlatLambdaCDM
    >>> from skypy.galaxy.spectrum import dirichlet_coefficients

    Calculate the flux density for two galaxies.

    >>> cosmology = FlatLambdaCDM(H0=70, Om0=0.3)
    >>> alpha0 = np.array([2.079, 3.524, 1.917, 1.992, 2.536])
    >>> alpha1 = np.array([2.265, 3.862, 1.921, 1.685, 2.480])
    >>> redshift = np.array([0.5,1])
    >>> coefficients = dirichlet_coefficients(redshift, alpha0, alpha1)
    >>> stellar_mass = np.array([5*10**10, 7*10**9])
    >>>
    >>> wavelength_o, sed = kcorrect_spectra(redshift, stellar_mass,
    ...                                       coefficients)

    """

    kcorrect_templates_url = "https://github.com/blanton144/kcorrect/raw/" \
                             "master/data/templates/k_nmf_derived.default.fits"
    templates = getdata(kcorrect_templates_url, 1)
    wavelength = getdata(kcorrect_templates_url, 11)

    rescaled_coeff = (coefficients.T * stellar_mass).T

    sed = (np.matmul(rescaled_coeff, templates).T / (1 + redshift)).T
    wavelength_observed = np.matmul((1 + redshift).reshape(len(redshift), 1),
                                    wavelength.reshape(1, len(wavelength)))

    return wavelength_observed, sed



import numpy as np
from astropy import constants, units
from skypy.galaxy.luminosity import herbel_luminosities
from skypy.gravitational_wave import b_band_merger_rate
from skypy.gravitational_wave.merger_rate import abadie_table_III


def test_abadie_rates():

    # Check the number of merger rates returned
    z, alpha, a_m, b_m = 1.0, -1.3, -0.9408582, -20.40492365
    luminosity = herbel_luminosities(z, alpha, a_m, b_m, size=1000)
    rates = b_band_merger_rate(luminosity, population='NS-NS', optimism='low')
    assert len(rates) == len(luminosity)

    # Test that a luminosity of L_10 (in units of solar luminosity) returns a
    # rate that matches the value in Abadie Table III
    L_10 = 1e10 * 2.16e33 / constants.L_sun.to_value('erg/s')
    L_B_rate = b_band_merger_rate(L_10, population='NS-NS', optimism='low')
    table_value = abadie_table_III['NS-NS']['low']
    assert np.isclose(L_B_rate.to_value(1/units.year), table_value, rtol=1e-5)



import numpy as np
import scipy.stats
import scipy.integrate
from skypy.utils import random, special


def test_schechter_cdf():
    # Test the _schechter_cdf function against precomputed values
    x = np.logspace(np.log10(0.1), np.log10(10), 10)
    cdf = random._schechter_cdf(x, min(x), max(x), -1.3)
    sample = np.array([0., 0.31938431, 0.57066007, 0.75733831, 0.88345969,
                       0.95631112, 0.98886695, 0.99846589, 0.99992254, 1.])

    np.testing.assert_allclose(cdf, sample)


def test_schechter():

    # Test the schechter function, sampling dimensionless x values
    alpha = -1.3
    x_min = 1e-10
    x_max = 1e2

    def calc_cdf(x):
        a = special.gammaincc(alpha + 1, x_min)
        b = special.gammaincc(alpha + 1, x)
        c = special.gammaincc(alpha + 1, x_min)
        d = special.gammaincc(alpha + 1, x_max)
        return (a - b) / (c - d)

    sample = random.schechter(alpha, x_min=x_min, x_max=x_max,
                              resolution=100, size=1000)

    # Test the distribution of galaxy properties follows the right distribution
    p_value = scipy.stats.kstest(sample, calc_cdf)[1]
    assert p_value > 0.01



"""Eisenstein and Hu.

This implements the Eisenstein and Hu fitting
formula for the matter power spectrum.
"""

from astropy.utils import isiterable
import numpy as np


__all__ = [
    'eisenstein_hu',
    'transfer_with_wiggles',
    'transfer_no_wiggles',
]


def transfer_with_wiggles(wavenumber, A_s, n_s, cosmology, kwmap=0.02):
    r''' Eisenstein & Hu transfer function with wiggles.
    This function returns the Eisenstein & Hu fitting formula for the transfer
    function with baryon acoustic oscillation wiggles. This is described in
    [1]_ and [2]_.

    Parameters
    ----------
    wavenumber : (nk,) array_like
        Array of wavenumbers in units of :math:`1/Mpc` at which to evaluate
        the linear matter power spectrum.
    cosmology : astropy.cosmology.Cosmology
        Cosmology object providing omega_matter, omega_baryon, Hubble parameter
        and CMB temperature at the present day.
    A_s, n_s: float
        Amplitude and spectral index of primordial scalar fluctuations.
    kwmap : float
        WMAP normalization for the amplitude of primordial scalar fluctuations,
        as described in [3]_, in units of :math:`1/Mpc`.
        Default is 0.02.

    Returns
    -------
    transfer : (nk,) array_like
        Transfer function evaluated at the given array of wavenumbers for the
        input primordial power spectrum parameters A_s and n_s, cosmology and
        kwmap normalization.

    Examples
    --------

    This returns the transfer function with wiggles for the Astropy default
    cosmology at a given array of wavenumbers:

    >>> import numpy as np
    >>> from astropy.cosmology import default_cosmology
    >>> wavenumber = np.logspace(-3, 1, num=5, base=10.0)
    >>> A_s, n_s = 2.1982e-09, 0.969453
    >>> cosmology = default_cosmology.get()
    >>> transfer_with_wiggles(wavenumber, A_s, n_s, cosmology, kwmap=0.02)
    array([9.92144790e-01, 7.78548704e-01, 1.29998169e-01, 4.63863054e-03,
       8.87918075e-05])

    References
    ----------
    .. [1] Eisenstein D. J., Hu W., ApJ, 496, 605 (1998)
    .. [2] Eisenstein D. J., Hu W., ApJ, 511, 5 (1999)
    .. [3] Komatsu et al., ApJS, 180, 330 (2009)

    '''

    if isiterable(wavenumber):
        wavenumber = np.asarray(wavenumber)
    if np.any(wavenumber <= 0):
        raise ValueError('Wavenumbers must be positive')

    om0 = cosmology.Om0
    ob0 = cosmology.Ob0
    h0 = cosmology.H0.value / 100
    Tcmb0 = cosmology.Tcmb0.value
    ak = wavenumber * h0
    om0h2 = om0 * h0**2
    ob0h2 = ob0 * h0**2
    f_baryon = ob0 / om0

    # redshift and wavenumber equality
    k_eq = 7.46e-2 * om0h2 * (Tcmb0 / 2.7)**-2
    z_eq = 2.5e4 * om0h2 * (Tcmb0 / 2.7)**-4

    # sound horizon and k_silk
    z_drag_b1 = 0.313 * om0h2**-0.419 * (1 + 0.607 * om0h2**0.674)
    z_drag_b2 = 0.238 * om0h2**0.223
    z_drag = 1291 * om0h2**0.251 / (1 + 0.659 * om0h2**0.828) \
        * (1 + z_drag_b1 * ob0h2**z_drag_b2)

    r_drag = 31.5 * ob0h2 * (Tcmb0 / 2.7)**-4 * (1000. / z_drag)
    r_eq = 31.5 * ob0h2 * (Tcmb0 / 2.7)**-4 * (1000. / z_eq)

    sound_horizon = 2 / (3 * k_eq) * np.sqrt(6 / r_eq) * \
        np.log((np.sqrt(1 + r_drag) + np.sqrt(r_drag + r_eq)) /
               (1 + np.sqrt(r_eq)))
    k_silk = 1.6 * ob0h2**0.52 * om0h2**0.73 * (1 + (10.4 * om0h2)**-0.95)

    # alpha c
    alpha_c_a1 = (46.9 * om0h2)**0.670 * (1 + (32.1 * om0h2)**-0.532)
    alpha_c_a2 = (12.0 * om0h2)**0.424 * (1 + (45.0 * om0h2)**-0.582)
    alpha_c = alpha_c_a1 ** -f_baryon * alpha_c_a2 ** (-f_baryon**3)

    # beta_c
    beta_c_b1 = 0.944 / (1 + (458 * om0h2)**-0.708)
    beta_c_b2 = (0.395 * om0h2)**-0.0266
    beta_c = 1 / (1 + beta_c_b1 * ((1 - f_baryon)**beta_c_b2 - 1))

    y = (1.0 + z_eq) / (1 + z_drag)
    alpha_b_G = y * (-6 * np.sqrt(1 + y) + (2 + 3 * y)
                     * np.log((np.sqrt(1 + y) + 1) / (np.sqrt(1 + y) - 1)))
    alpha_b = 2.07 * k_eq * sound_horizon * (1 + r_drag)**-0.75 * alpha_b_G

    beta_node = 8.41 * om0h2 ** 0.435
    beta_b = 0.5 + f_baryon + (3 - 2 * f_baryon) * np.sqrt((17.2 * om0h2)**2
                                                           + 1.0)

    q = ak / (13.41 * k_eq)
    ks = ak * sound_horizon

    T_c_ln_beta = np.log(np.e + 1.8 * beta_c * q)
    T_c_ln_nobeta = np.log(np.e + 1.8 * q)
    T_c_C_alpha = 14.2 / alpha_c + 386. / (1 + 69.9 * q ** 1.08)
    T_c_C_noalpha = 14.2 + 386. / (1 + 69.9 * q ** 1.08)

    T_c_f = 1 / (1 + (ks / 5.4) ** 4)

    def f(a, b):
        return a / (a + b * q**2)

    T_c = T_c_f * f(T_c_ln_beta, T_c_C_noalpha) + \
        (1 - T_c_f) * f(T_c_ln_beta, T_c_C_alpha)

    s_tilde = sound_horizon * (1 + (beta_node / ks)**3)**(-1 / 3)
    ks_tilde = ak * s_tilde

    T_b_T0 = f(T_c_ln_nobeta, T_c_C_noalpha)
    T_b_1 = T_b_T0 / (1 + (ks / 5.2)**2)
    T_b_2 = alpha_b / (1 + (beta_b / ks)**3) * np.exp(-(ak / k_silk)**1.4)
    T_b = np.sinc(ks_tilde / np.pi) * (T_b_1 + T_b_2)

    transfer = f_baryon * T_b + (1 - f_baryon) * T_c

    return transfer


def transfer_no_wiggles(wavenumber, A_s, n_s, cosmology):
    r'''Eisenstein & Hu transfer function without wiggles.
    Eisenstein & Hu fitting formula for the transfer function without
    baryon acoustic oscillation wiggles. This is described in
    [1]_ and [2]_.

    Parameters
    ----------
    wavenumber : (nk,) array_like
        Array of wavenumbers in units of :math:`1/Mpc` at which to evaluate
        the linear matter power spectrum.
    A_s, n_s: float
        Amplitude and spectral index of primordial scalar fluctuations.
    cosmology : astropy.cosmology.Cosmology
        Cosmology object providing omega_matter, omega_baryon, Hubble parameter
        and CMB temperature in the present day.

    Returns
    -------
    transfer : (nk, ) array_like
        Transfer function evaluated at the given wavenumbers for the input
        primordial power spectrum parameters A_s and n_s, cosmology and kwmap
        normalization.

    Examples
    --------

    This returns the transfer function without wiggles for the Astropy default
    cosmology at a given array of wavenumbers:

    >>> import numpy as np
    >>> from astropy.cosmology import default_cosmology
    >>> wavenumber = np.logspace(-3, 1, num=5, base=10.0)
    >>> A_s, n_s = 2.1982e-09, 0.969453
    >>> cosmology = default_cosmology.get()
    >>> transfer_no_wiggles(wavenumber, A_s, n_s, cosmology)
    array([9.91959695e-01, 7.84518347e-01, 1.32327555e-01, 4.60773671e-03,
       8.78447096e-05])

    References
    ----------
    .. [1] Eisenstein D. J., Hu W., ApJ, 496, 605 (1998)
    .. [2] Eisenstein D. J., Hu W., ApJ, 511, 5 (1999)
    .. [3] Komatsu et al., ApJS, 180, 330 (2009)

    '''

    if isiterable(wavenumber):
        wavenumber = np.asarray(wavenumber)
    if np.any(wavenumber <= 0):
        raise ValueError('Wavenumbers must be positive')

    om0 = cosmology.Om0
    ob0 = cosmology.Ob0
    h0 = cosmology.H0.value / 100
    Tcmb0 = cosmology.Tcmb0.value
    ak = wavenumber * h0
    om0h2 = om0 * h0**2
    f_baryon = ob0 / om0

    alpha = 1 - 0.328 * np.log(431 * om0h2) * f_baryon + 0.38 * \
        np.log(22.3 * om0h2) * f_baryon**2
    sound = 44.5 * np.log(9.83 / om0h2) / \
        np.sqrt(1 + 10 * (f_baryon * om0h2)**(0.75))
    shape = om0h2 * (alpha + (1 - alpha) / (1 + (0.43 * ak * sound)**4))
    aq = ak * (Tcmb0 / 2.7)**2 / shape
    transfer = np.log(2 * np.e + 1.8 * aq) / \
        (np.log(2 * np.e + 1.8 * aq) +
         (14.2 + 731 / (1 + 62.5 * aq)) * aq * aq)

    return transfer


def eisenstein_hu(wavenumber, A_s, n_s, cosmology, kwmap=0.02, wiggle=True):
    """ Eisenstein & Hu matter power spectrum.
    This function returns the Eisenstein and Hu fitting function for the linear
    matter power spectrum with (or without) baryon acoustic oscillations, c.f.
    [1]_ and [2]_, using
    formulation from Komatsu et al (2009) in [3]_.

    Parameters
    ----------
    wavenumber : (nk, ) array_like
        Array of wavenumbers in units of :math:`1/Mpc` at which to evaluate
        the linear matter power spectrum.
    A_s, n_s: float
        Amplitude and spectral index of primordial scalar fluctuations.
    cosmology : astropy.cosmology.Cosmology
        Cosmology object providing omega_matter, omega_baryon, Hubble parameter
        and CMB temperature in the present day.
    kwmap : float
        WMAP normalization for the amplitude of primordial scalar fluctuations,
        as described in [3], in units of :math:`1/Mpc`. Default is 0.02.
    wiggle : bool
        Boolean flag to set the use of baryion acoustic oscillations wiggles.
        Default is True, for which the power spectrum is computed with the
        wiggles.

    Returns
    -------
    power_spectrum : array_like
        Linear matter power spectrum in units of :math:`Mpc^3`,
        evaluated at the given wavenumbers for the input primordial
        power spectrum parameters
        A_s and n_s, cosmology, and kwmap normalization.

    Examples
    --------

    This example returns the Eisenstein and Hu matter power spectrum with
    baryon acoustic oscillations for the Astropy default
    cosmology at a given array of wavenumbers:

    >>> import numpy as np
    >>> from astropy.cosmology import default_cosmology
    >>> wavenumber = np.logspace(-3, 1, num=5, base=10.0)
    >>> A_s, n_s = 2.1982e-09, 0.969453
    >>> cosmology = default_cosmology.get()
    >>> eisenstein_hu(wavenumber, A_s, n_s, cosmology, kwmap=0.02,
    ...               wiggle=True)
    array([6.47460158e+03, 3.71610099e+04, 9.65702614e+03, 1.14604456e+02,
       3.91399918e-01])

    References
    ----------
    .. [1] Eisenstein D. J., Hu W., ApJ, 496, 605 (1998)
    .. [2] Eisenstein D. J., Hu W., ApJ, 511, 5 (1999)
    .. [3] Komatsu et al., ApJS, 180, 330 (2009)
    """
    om0 = cosmology.Om0
    h0 = cosmology.H0.value / 100
    if wiggle:
        transfer = transfer_with_wiggles(wavenumber, A_s, n_s, cosmology,
                                         kwmap)
    else:
        transfer = transfer_no_wiggles(wavenumber, A_s, n_s, cosmology)

    # Eq [74] in [3]
    power_spectrum = A_s * (2 * wavenumber**2 * 2998**2 / 5 / om0)**2 * \
        transfer**2 * (wavenumber * h0 / kwmap)**(n_s - 1) * 2 * \
        np.pi**2 / wavenumber**3

    return power_spectrum



"""Growth function.

This computes the linear growth function in
perturbation theory.
"""

from astropy.utils import isiterable
import numpy as np
from scipy import integrate


__all__ = [
   'growth_factor',
   'growth_function',
   'growth_function_carroll',
   'growth_function_derivative',
]


def growth_function_carroll(redshift, cosmology):
    '''Growth function.

    This function returns the growth function as a function of redshift for a
    given cosmology as approximated by Carroll, Press & Turner (1992),
    equation 29 in [1]_.

    Parameters
    ----------
    redshift : (nz,) array_like
        Array of redshifts at which to evaluate the growth function.
    cosmology : astropy.cosmology.Cosmology
        Cosmology object providing methods for the evolution history of
        omega_matter and omega_lambda with redshift.

    Returns
    -------
    growth : (nz,) array_like
        The growth function evaluated at the input redshifts for the given
        cosmology.

    Examples
    --------

    This example returns the growth function for a given array of redshifts
    and for the Astropy default cosmology:

    >>> import numpy as np
    >>> from astropy.cosmology import default_cosmology
    >>> redshift = np.array([0, 1, 2])
    >>> cosmology = default_cosmology.get()
    >>> growth_function_carroll(redshift, cosmology)
    array([0.781361..., 0.476280..., 0.327549...])

    References
    ----------
    .. [1] Carroll, M. and Press, W. and Turner, E., (1992),
        doi : 10.1146/annurev.aa.30.090192.002435
    '''

    if isiterable(redshift):
        redshift = np.asarray(redshift)
    if np.any(redshift < 0):
        raise ValueError('Redshifts must be non-negative')

    Om = cosmology.Om(redshift)
    Ode = cosmology.Ode(redshift)
    Dz = 2.5 * Om / (1 + redshift)
    return Dz / (np.power(Om, 4.0/7.0) - Ode + (1 + 0.5*Om) * (1.0 + Ode/70.0))


def growth_factor(redshift, cosmology, gamma=6.0/11.0):
    r'''Growth factor.

    Function used to calculate :math:`f(z)`, parametrised growth factor as a
    function of redshift, as described in [1]_ equation 17.

    Parameters
    ----------
    redshift : (nz,) array_like
        Array of redshifts at which to evaluate the growth function.
    cosmology : astropy.cosmology.Cosmology
        Cosmology object providing methods for the evolution history of
        omega_matter and omega_lambda with redshift.
    gamma : float
        Growth index providing an efficient parametrization of the matter
        perturbations.

    Returns
    -------
    growth_factor : (nz,) array_like
      The redshift scaling of the growth factor.

    Examples
    --------

    This example returns the growth factor for a given array of redshifts
    and for a given cosmology:

    >>> import numpy as np
    >>> from astropy.cosmology import FlatLambdaCDM
    >>> cosmology = FlatLambdaCDM(H0=67.04, Om0=0.3183, Ob0=0.047745)
    >>> growth_factor(0, cosmology)
    0.5355746155304598

    References
    ----------
    .. [1] E. V. Linder, Phys. Rev. D 72, 043529 (2005)
    '''
    z = redshift

    omega_m_z = cosmology.Om(z)
    growth_factor = np.power(omega_m_z, gamma)

    return growth_factor


def growth_function(redshift, cosmology, gamma=6.0/11.0):
    r'''Growth function.

    Function used to calculate :math:`D(z)`, growth function at different
    redshifts, as described in [1]_ equation 16.

    Parameters
    ----------
    redshift : (nz,) array_like
        Array of redshifts at which to evaluate the growth function.
    cosmology : astropy.cosmology.Cosmology
        Cosmology object providing methods for the evolution history of
        omega_matter and omega_lambda with redshift.
    gamma : float
        Growth index providing an efficient parametrization of the matter
        perturbations.

    Returns
    -------
    growth_function : (nz,) array_like
      The redshift scaling of the growth function.

    Examples
    --------

    This example returns the growth function for a given array of redshifts
    and for a given cosmology:

    >>> import numpy as np
    >>> from scipy import integrate
    >>> from astropy.cosmology import FlatLambdaCDM
    >>> cosmology = FlatLambdaCDM(H0=67.04, Om0=0.3183, Ob0=0.047745)
    >>> growth_function(0, cosmology)
    0.7909271056297236

    References
    ----------
    .. [1] E. V. Linder, Phys. Rev. D 72, 043529 (2005)
    '''
    z = redshift

    def integrand(x):
        integrand = (growth_factor(x, cosmology, gamma) - 1) / (1 + x)
        return integrand

    if isinstance(z, int) or isinstance(z, float):
        integral = integrate.quad(integrand, z, 1100)[0]
        g = np.exp(integral)
        growth_function = g / (1 + z)

    elif isinstance(z, np.ndarray):
        growth_function = np.zeros(np.shape(z))

        for i, aux in enumerate(z):
            integral = integrate.quad(integrand, aux, 1100)[0]
            g = np.exp(integral)
            growth_function[i] = g / (1 + aux)

    else:
        raise ValueError('Redshift must be integer, float or ndarray ')

    return growth_function


def growth_function_derivative(redshift, cosmology, gamma=6.0/11.0):
    r'''First derivative of the growth function.

    Function used to calculate D'(z), derivative of the growth function
    with respect to redshift as in [1]_ equation 16.

    Parameters
    ----------
    redshift : (nz,) array_like
        Array of redshifts at which to evaluate the growth function.
    cosmology : astropy.cosmology.Cosmology
        Cosmology object providing methods for the evolution history of
        omega_matter and omega_lambda with redshift.
    gamma : float
        Growth index providing an efficient parametrization of the matter
        perturbations.

    Returns
    -------
    growth_function_derivative : (nz,) array_like
      The redshift scaling of the derivative of the growth function.

    Notes
    -----
    The first derivative of the growth function, :math:`D(z)`,
    with respect to redshift reads

    .. math:: D'(z) = - \frac{D(z) f(z)}{1 + z} \;.

    With :math:`f(z)` the growth factor.


    Examples
    --------

    This example returns the first derivative of the growth function for a
    given array of redshifts and cosmology:

    >>> import numpy as np
    >>> from scipy import integrate
    >>> from astropy.cosmology import FlatLambdaCDM
    >>> cosmology = FlatLambdaCDM(H0=67.04, Om0=0.3183, Ob0=0.047745)
    >>> growth_function_derivative(0, cosmology)
    -0.42360048051025856

    References
    ----------
    .. [1] E. V. Linder, Phys. Rev. D 72, 043529 (2005)
    '''
    z = redshift

    growth_function_derivative = - growth_function(z, cosmology, gamma) * \
        growth_factor(z, cosmology, gamma) / (1.0 + z)

    return growth_function_derivative



r'''Halo mass sampler.
This code samples halos from their mass function.

Models
======
.. autosummary::
   :nosignatures:
   :toctree: ../api/

   press_schechter
   halo_mass_function
   halo_mass_sampler
   ellipsoidal_collapse_function
   press_schechter_collapse_function
   sheth_tormen_collapse_function
   press_schechter
   sheth_tormen
'''

import numpy as np
from scipy import integrate
from functools import partial
from astropy import units

__all__ = [
    'halo_mass_function',
    'halo_mass_sampler',
    'sheth_tormen_collapse_function',
    'press_schechter_collapse_function',
    'sheth_tormen_mass_function',
    'press_schechter_mass_function',
    'sheth_tormen',
    'press_schechter',
 ]


def halo_mass_function(M, wavenumber, power_spectrum, growth_function,
                       cosmology, collapse_function, params):
    r'''Halo mass function.
    This function computes the halo mass function, defined
    in equation 7.46 in [1]_.

    Parameters
    -----------
    M : (nm,) array_like
        Array for the halo mass, in units of solar mass.
    wavenumber : (nk,) array_like
        Array of wavenumbers at which the power spectrum is evaluated,
        in units of :math:`1/Mpc`.
    power_spectrum: (nk,) array_like
        Linear power spectrum at redshift 0 in :math:`Mpc^3`.
    growth_function : float
        The growth function evaluated at a given redshift for the given
        cosmology.
    cosmology : astropy.cosmology.Cosmology
        Cosmology object providing methods for the evolution history of
        omega_matter and omega_lambda with redshift.
    collapse_function: function
        Collapse function to choose from a variety of models:
        `sheth_tormen_collapse_function`, `press_schechter_collapse_function`.
    params: tuple
        List of parameters that determines the model used for
        the collapse function.

    Returns
    --------
    mass_function: (nm,) array_like
        Halo mass function for a given mass array, cosmology and redshift, in
        units of :math:`Mpc^{-3} M_{Sun}^{-1}`.

    Examples
    ---------
    >>> import numpy as np
    >>> from skypy.halo import mass
    >>> from skypy.power_spectrum import eisenstein_hu

    This example will compute the halo mass function for elliptical and
    spherical collapse, for a Planck15 cosmology at redshift 0.
    The power spectrum is given by the Eisenstein and Hu fitting formula:

    >>> from astropy.cosmology import Planck15
    >>> cosmo = Planck15
    >>> D0 = 1.0
    >>> k = np.logspace(-3, 1, num=1000, base=10.0)
    >>> A_s, n_s = 2.1982e-09, 0.969453
    >>> Pk = eisenstein_hu(k, A_s, n_s, cosmo, kwmap=0.02, wiggle=True)

    The Sheth and Tormen mass function at redshift 0:

    >>> m = 10**np.arange(9.0, 12.0, 2)
    >>> mass.sheth_tormen_mass_function(m, k, Pk, D0, cosmo)
    array([3.07523240e-11, 6.11387743e-13])

    And the Press-Schechter mass function at redshift 0:

    >>> mass.press_schechter_mass_function(m, k, Pk, D0, cosmo)
    array([3.46908809e-11, 8.09874945e-13])

    For any other collapse models:

    >>> params_model = (0.3, 0.7, 0.3, 1.686)
    >>> mass.halo_mass_function(m, k, Pk, D0, cosmo,
    ...     ellipsoidal_collapse_function, params=params_model)
    array([2.85598921e-11, 5.67987501e-13])

    References
    ----------
    .. [1] Mo, H. and van den Bosch, F. and White, S. (2010), Cambridge
        University Press, ISBN: 9780521857932.
    '''
    sigma = np.sqrt(_sigma_squared(M, wavenumber, power_spectrum,
                                   growth_function, cosmology))
    f_c = collapse_function(sigma, params)

    dlognu_dlogm = _dlns_dlnM(sigma, M)
    rho_bar = (cosmology.critical_density0.to(units.Msun / units.Mpc ** 3)).value
    rho_m0 = cosmology.Om0 * rho_bar

    return rho_m0 * f_c * dlognu_dlogm / np.power(M, 2)


def halo_mass_sampler(m_min, m_max, resolution, wavenumber, power_spectrum,
                      growth_function, cosmology,
                      collapse_function, params, size=None):
    r'''Halo mass sampler.
    This function samples haloes from their mass function,
    see equation 7.46 in [1]_.

    Parameters
    -----------
    m_min, m_max : array_like
        Lower and upper bounds for the random variable m.
    resolution: int
        Resolution of the inverse transform sampling spline.
    wavenumber : (nk,) array_like
        Array of wavenumbers at which the power spectrum is evaluated,
        in units of :math:`1/Mpc`.
    power_spectrum: (nk,) array_like
        Linear power spectrum at redshift 0 in :math:`Mpc^3`.
    growth_function : float
        The growth function evaluated at a given redshift for the given
        cosmology.
    cosmology : astropy.cosmology.Cosmology
        Cosmology object providing methods for the evolution history of
        omega_matter and omega_lambda with redshift.
    collapse_function: function
        Collapse function to choose from a variety of models:
        `sheth_tormen_collapse_function`, `press_schechter_collapse_function`.
    params: tuple
        List of parameters that determines the model used for
        the collapse function.
    size: int, optional
        Output shape of samples. Default is None.

    Returns
    --------
    sample: (size,) array_like
        Samples drawn from the mass function, in units of solar masses.

    Examples
    ---------
    >>> import numpy as np
    >>> from skypy.halo import mass
    >>> from skypy.power_spectrum import eisenstein_hu

    This example will sample from the halo mass function for
    a Planck15 cosmology at redshift 0. The power spectrum is given
    by the Eisenstein and Hu fitting formula:

    >>> from astropy.cosmology import Planck15
    >>> cosmo = Planck15
    >>> D0 = 1.0
    >>> k = np.logspace(-3, 1, num=100, base=10.0)
    >>> A_s, n_s = 2.1982e-09, 0.969453
    >>> Pk = eisenstein_hu(k, A_s, n_s, cosmo, kwmap=0.02, wiggle=True)

    Sampling from the Sheth and Tormen mass function:

    >>> halo_mass = mass.sheth_tormen(1e9, 1e12, 100, k, Pk, D0, cosmo)

    And from the Press-Schechter mass function:

    >>> halo_mass = mass.press_schechter(1e9, 1e12, 100, k, Pk, D0, cosmo)

    For any other collapse models:

    >>> params_model = (0.3, 0.7, 0.3, 1.686)
    >>> halo_mass = mass.halo_mass_sampler(1e9, 1e12, 100, k, Pk, D0, cosmo,
    ...     ellipsoidal_collapse_function, params=params_model)

    References
    ----------
    .. [1] Mo, H. and van den Bosch, F. and White, S. (2010), Cambridge
        University Press, ISBN: 9780521857932.
    '''
    m = np.logspace(np.log10(m_min), np.log10(m_max), resolution)

    massf = halo_mass_function(
            m, wavenumber, power_spectrum, growth_function,
            cosmology, collapse_function, params=params)

    CDF = integrate.cumtrapz(massf, m, initial=0)
    CDF = CDF / CDF[-1]
    n_uniform = np.random.uniform(size=size)
    return np.interp(n_uniform, CDF, m)


def ellipsoidal_collapse_function(sigma, params):
    r'''Ellipsoidal collapse function.
    This function computes the mass function for ellipsoidal
    collapse, see equation 10 in [1]_ or [2]_.

    Parameters
    -----------
    sigma: (ns,) array_like
        Array of the mass variance at different scales and at a given redshift.
    params: float
        The :math:`{A,a,p, delta_c}` parameters of the Sheth-Tormen formalism.

    Returns
    --------
    f_c: array_like
        Array with the values of the collapse function.

    Examples
    ---------
    >>> import numpy as np
    >>> from skypy.halo import mass
    >>> from skypy.power_spectrum import eisenstein_hu
    >>> from skypy.power_spectrum import growth_function

    This example will compute the mass function for
    ellipsoidal collapse and a Planck15 cosmology at redshift 0.
    The power spectrum is given by the Eisenstein and Hu fitting formula:

    >>> from astropy.cosmology import Planck15
    >>> cosmo = Planck15
    >>> D0 = 1.0
    >>> k = np.logspace(-3, 1, num=5, base=10.0)
    >>> A_s, n_s = 2.1982e-09, 0.969453
    >>> Pk = eisenstein_hu(k, A_s, n_s, cosmo, kwmap=0.02, wiggle=True)

    The Sheth-Tormen collapse function at redshift 0:

    >>> m = 10**np.arange(9.0, 12.0, 2)
    >>> sigma = np.sqrt(_sigma_squared(m, k, Pk, D0, cosmo))
    >>> mass.sheth_tormen_collapse_function(sigma)
    array([0.17947815, 0.19952375])

    And the Press-Schechter collapse function at redshift 0:

    >>> mass.press_schechter_collapse_function(sigma)
    array([0.17896132, 0.21613726])

    For any other collapse models:

    >>> params_model = (0.3, 0.7, 0.3, 1.686)
    >>> mass.ellipsoidal_collapse_function(sigma, params=params_model)
    array([0.16667541, 0.18529452])

    References
    ----------
    .. [1] R. K. Sheth and G. Tormen,  Mon. Not. Roy. Astron. Soc. 308, 119
        (1999), astro-ph/9901122.
    .. [2] https://www.slac.stanford.edu/econf/C070730/talks/
        Wechsler_080207.pdf
    '''
    A, a, p, delta_c = params

    return A * np.sqrt(2.0 * a / np.pi) * (delta_c / sigma) * \
        np.exp(- 0.5 * a * (delta_c / sigma)**2) * \
        (1.0 + np.power(np.power(sigma / delta_c, 2.0) / a, p))


press_schechter_collapse_function = partial(ellipsoidal_collapse_function,
                                            params=(0.5, 1, 0, 1.69))
sheth_tormen_collapse_function = partial(ellipsoidal_collapse_function,
                                         params=(0.3222, 0.707, 0.3, 1.686))
sheth_tormen_mass_function = partial(
                             halo_mass_function,
                             collapse_function=ellipsoidal_collapse_function,
                             params=(0.3222, 0.707, 0.3, 1.686))
press_schechter_mass_function = partial(
                                halo_mass_function,
                                collapse_function=ellipsoidal_collapse_function,
                                params=(0.5, 1, 0, 1.69))
sheth_tormen = partial(halo_mass_sampler,
                       collapse_function=ellipsoidal_collapse_function,
                       params=(0.3222, 0.707, 0.3, 1.686))
press_schechter = partial(halo_mass_sampler,
                          collapse_function=ellipsoidal_collapse_function,
                          params=(0.5, 1, 0, 1.69))


def _sigma_squared(M, k, Pk, growth_function, cosmology):
    M = np.atleast_1d(M)[:, np.newaxis]

    # Growth function
    Dz2 = np.power(growth_function, 2)

    # Matter mean density today
    rho_bar = (cosmology.critical_density0.to(units.Msun / units.Mpc ** 3)).value
    rho_m0 = cosmology.Om0 * rho_bar

    R = np.power(3 * M / (4 * np.pi * rho_m0), 1.0 / 3.0)
    top_hat = 3. * (np.sin(k * R) - k * R * np.cos(k * R)) / ((k * R)**3.)
    integrand = Pk * np.power(top_hat * k, 2)

    return Dz2 * integrate.simps(integrand, k) / (2. * np.pi**2.)


def _dlns_dlnM(sigma, M):
    ds = np.gradient(sigma, M)
    return np.absolute((M / sigma) * ds)



import numpy as np
from scipy import stats


def test_beta_ellipticity():

    from skypy.galaxy.ellipticity import beta_ellipticity

    # randomised ellipticity distribution with beta distribution parameters a,b
    # and the equivalent reparametrisation
    a, b = np.random.lognormal(size=2)
    e_ratio, e_sum = (a / (a + b), a + b)

    # Test scalar output
    assert np.isscalar(beta_ellipticity(e_ratio, e_sum))

    # Test array output
    assert beta_ellipticity(e_ratio, e_sum, size=10).shape == (10,)

    # Test broadcast output
    e_ratio2 = 0.5 * np.ones((13, 1, 5))
    e_sum2 = 0.5 * np.ones((7, 5))
    rvs = beta_ellipticity(e_ratio2, e_sum2)
    assert rvs.shape == np.broadcast(e_ratio2, e_sum2).shape

    # Kolmogorov-Smirnov test comparing ellipticity and beta distributions
    D, p = stats.kstest(beta_ellipticity(e_ratio, e_sum, size=1000), 'beta',
                        args=(a, b))
    assert p > 0.01, 'D = {}, p = {}'.format(D, p)

    # Kolmogorov-Smirnov test comparing ellipticity and uniform distributions
    D, p = stats.kstest(beta_ellipticity(0.5, 2.0, size=1000), 'uniform')
    assert p > 0.01, 'D = {}, p = {}'.format(D, p)

    # Kolmogorov-Smirnov test comparing ellipticity and arcsine distributions
    D, p = stats.kstest(beta_ellipticity(0.5, 1.0, size=1000), 'arcsine')
    assert p > 0.01, 'D = {}, p = {}'.format(D, p)


def test_ryden04():
    from skypy.galaxy.ellipticity import ryden04

    # sample a single ellipticity
    e = ryden04(0.222, 0.056, -1.85, 0.89)
    assert np.isscalar(e)

    # sample many ellipticities
    e = ryden04(0.222, 0.056, -1.85, 0.89, size=1000)
    assert np.shape(e) == (1000,)

    # sample with explicit shape
    e = ryden04(0.222, 0.056, -1.85, 0.89, size=(10, 10))
    assert np.shape(e) == (10, 10)

    # sample with implicit size
    e1 = ryden04([0.222, 0.333], 0.056, -1.85, 0.89)
    e2 = ryden04(0.222, [0.056, 0.067], -1.85, 0.89)
    e3 = ryden04(0.222, 0.056, [-1.85, -2.85], 0.89)
    e4 = ryden04(0.222, 0.056, -1.85, [0.89, 1.001])
    assert np.shape(e1) == np.shape(e2) == np.shape(e3) == np.shape(e4) == (2,)

    # sample with broadcasting rule
    e = ryden04([[0.2, 0.3], [0.4, 0.5]], 0.1, [-1.9, -2.9], 0.9)
    assert np.shape(e) == (2, 2)

    # sample with random parameters and check that result is in unit range
    args = np.random.rand(4)*[1., .1, -2., 1.]
    e = ryden04(*args, size=1000)
    assert np.all((e >= 0.) & (e <= 1.))

    # sample a spherical distribution
    e = ryden04(1-1e-99, 1e-99, -1e99, 1e-99, size=1000)
    assert np.allclose(e, 0.)



from astropy.utils import isiterable
from collections import namedtuple
from functools import partial
import numpy as np
from scipy import interpolate
from scipy import integrate
from scipy import optimize


__all__ = [
   'HalofitParameters',
   'halofit',
   'halofit_smith',
   'halofit_takahashi',
   'halofit_bird',
]


HalofitParameters = namedtuple(
    'HalofitParameters',
    ['a', 'b', 'c', 'gamma', 'alpha', 'beta', 'mu', 'nu', 'fa', 'fb',
     'l', 'm', 'p', 'r', 's', 't'])

_smith_parameters = HalofitParameters(
    [0.1670, 0.7940, 1.6762, 1.8369, 1.4861, -0.6206, 0.0],
    [0.3084, 0.9466, 0.9463, -0.9400, 0.0],
    [0.3214, 0.6669, -0.2807, -0.0793],
    [0.2989, 0.8649, 0.1631],
    [-0.1452, 0.3700, 1.3884, 0.0],
    [0.0, 0.0, 0.3401, 0.9854, 0.8291, 0.0],
    [0.1908, -3.5442],
    [1.2857, 0.9589],
    [-0.0732, -0.1423, 0.0725],
    [-0.0307, -0.0585, 0.0743],
    0.0, 0.0, 0.0, 0.0, 0.0, 0.0)

_takahashi_parameters = HalofitParameters(
    [0.2250, 0.9903, 2.3706, 2.8553, 1.5222, -0.6038, 0.1749],
    [0.5716, 0.5864, -0.5642, -1.5474, 0.2279],
    [0.8161, 2.0404, 0.3698, 0.5869],
    [-0.0843, 0.1971, 0.8460],
    [-0.1959, 1.3373, 6.0835, -5.5274],
    [0.3980, 1.2490, 0.3157, -0.7354, 2.0379, -0.1682],
    [0.0, -np.inf],
    [3.6902, 5.2105],
    [-0.0732, -0.1423, 0.0725],
    [-0.0307, -0.0585, 0.0743],
    0.0, 0.0, 0.0, 0.0, 0.0, 0.0)

_bird_parameters = HalofitParameters(
    [0.1670, 0.7940, 1.6762, 1.8369, 1.4861, -0.6206, 0.0],
    [0.3084, 0.9466, 0.9463, -0.9400, 0.0],
    [0.3214, 0.6669, -0.2807, -0.0793],
    [0.2224, 1.18075, -0.6719],
    [-0.1452, 0.3700, 1.3884, 0.0],
    [0.0, 0.0, 0.3401, 0.9854, 0.8291, 0.0],
    [0.1908, -3.5442],
    [1.2857, 0.9589],
    [-0.0732, -0.1423, 0.0725],
    [-0.0307, -0.0585, 0.0743],
    2.080, 1.2e-3, 26.3, -6.49, 1.44, 12.4)


def halofit(wavenumber, redshift, linear_power_spectrum,
            cosmology, parameters):
    r'''Computation of the non-linear halo power spectrum.

    This function computes the non-linear halo power spectrum, as a function
    of redshift and wavenumbers, following [1]_, [2]_ and [3]_.

    Parameters
    ----------
    k : (nk,) array_like
        Input wavenumbers in units of :math:`1/Mpc`.
    z : (nz,) array_like
        Input redshifts
    P : (nk, nz) array_like
        Linear power spectrum for given wavenumbers
        and redshifts :math:`Mpc^3`.
    cosmology : astropy.cosmology.Cosmology
                Cosmology object providing method for the evolution of
                omega_matter with redshift.
    parameters : HalofitParameters
                 namedtuple containing the free parameters of the model.

    Returns
    -------
    pknl : (nk, nz) array_like
           Non-linear halo power spectrum in units of :math:`Mpc^3`.

    References
    ----------
    .. [1] R. E. Smith it et al., VIRGO Consortium,
           Mon. Not. Roy. Astron. Soc. 341, 1311 (2003).
    .. [2] R. Takahashi, M. Sato, T. Nishimichi, A. Taruya and M. Oguri,
           Astrophys. J. 761, 152 (2012).
    .. [3] S. Bird, M. Viel and M. G. Haehnelt,
           Mon. Not. Roy. Astron. Soc. 420, 2551 (2012).

    Examples
    --------
    >>> import numpy as np
    >>> from astropy.cosmology import FlatLambdaCDM
    >>> kvec = np.array([1.00000000e-04, 1.01000000e+01])
    >>> zvalue = 0.0
    >>> pvec = np.array([388.6725682632502, 0.21676249605280398])
    >>> cosmo = FlatLambdaCDM(H0=67.04, Om0=0.21479, Ob0=0.04895)
    >>> halofit(kvec, zvalue, pvec, cosmo, _takahashi_parameters)
    array([388.67064424,   0.72797614])
    '''

    # Manage shapes of input arrays
    return_shape = np.shape(linear_power_spectrum)
    redshift = np.atleast_1d(redshift)
    if np.ndim(linear_power_spectrum) == 1:
        linear_power_spectrum = linear_power_spectrum[:, np.newaxis]

    # Declaration of variables
    if isiterable(redshift):
        redshift = np.asarray(redshift)
    if isiterable(wavenumber):
        wavenumber = np.asarray(wavenumber)
    if isiterable(linear_power_spectrum):
        linear_power_spectrum = np.asarray(linear_power_spectrum)
    if np.any(redshift < 0):
        raise ValueError('Redshifts must be non-negative')
    if np.any(wavenumber <= 0):
        raise ValueError('Wavenumbers must be strictly positive')
    if np.any(linear_power_spectrum < 0):
        raise ValueError('Linear power spectrum must be non-negative')
    if not np.all(sorted(wavenumber) == wavenumber):
        raise ValueError('Wavenumbers must be provided in ascending order')

    # Redshift-dependent quantities from cosmology
    omega_m_z = cosmology.Om(redshift)[:, np.newaxis]
    omega_nu_z = cosmology.Onu(redshift)[:, np.newaxis]
    omega_de_z = cosmology.Ode(redshift)[:, np.newaxis]
    wp1_z = 1.0 + cosmology.w(redshift)[:, np.newaxis]
    ode_1pw_z = omega_de_z * wp1_z

    # Linear power spectrum interpolated at each redshift
    k2 = np.square(wavenumber)
    k3 = np.power(wavenumber, 3)
    dl2kz = (linear_power_spectrum.T * k3) / (2 * np.pi * np.pi)
    dl2k = [interpolate.interp1d(np.log(wavenumber), np.log(d)) for d in dl2kz]
    lnk_lo = np.log(wavenumber[0])
    lnk_up = np.log(wavenumber[-1])

    # Integrals required to evaluate Smith et al. 2003 equations C5, C7 & C8
    def integrand_kn(lnk, lnR, lnd, n):
        R2 = np.exp(2*lnR)
        k2 = np.exp(2*lnk)
        dl2 = np.exp(lnd(lnk))
        return dl2 * np.power(k2, n/2) * np.exp(-k2*R2)

    def integral_kn(lnR, lnd, n, lnk_lo, lnk_up):
        integrand = partial(integrand_kn, lnR=lnR, lnd=lnd, n=n)
        return integrate.quad(integrand, lnk_lo, lnk_up)[0]

    # Find root at which sigma^2(R) == 1.0 for each redshift
    # Smith et al. 2003 equation C5 & C6
    def log_sigma_squared(lnR):
        ik0 = [integral_kn(r, d, 0, lnk_lo, lnk_up) for r, d in zip(lnR, dl2k)]
        return np.log(ik0)
    guess = np.zeros(np.size(redshift))
    root = optimize.fsolve(log_sigma_squared, guess)
    R = np.exp(root)[:, np.newaxis]
    ksigma = 1.0 / R
    y = wavenumber / ksigma

    # Evaluate integrals at lnR = root for each redshift
    ik0 = [integral_kn(r, d, 0, lnk_lo, lnk_up) for r, d in zip(root, dl2k)]
    ik2 = [integral_kn(r, d, 2, lnk_lo, lnk_up) for r, d in zip(root, dl2k)]
    ik4 = [integral_kn(r, d, 4, lnk_lo, lnk_up) for r, d in zip(root, dl2k)]
    ik0 = np.asarray(ik0)[:, np.newaxis]
    ik2 = np.asarray(ik2)[:, np.newaxis]
    ik4 = np.asarray(ik4)[:, np.newaxis]

    # Effective spectral index neff and curvature C
    # Smith et al. 2003 equations C7 & C8
    neff = (2 * R * R * ik2 / ik0) - 3
    c = (4 * R * R / ik0) * (ik2 + R * R * (ik2 * ik2 / ik0 - ik4))

    # Smith et al. 2003 equations C9-C16
    # With higher order terms from Takahashi et al. 2012 equations A6-A13
    p = parameters
    an = np.power(10, np.polyval(p.a[:5], neff) + p.a[5]*c + p.a[6]*ode_1pw_z)
    bn = np.power(10, np.polyval(p.b[:3], neff) + p.b[3]*c + p.a[4]*ode_1pw_z)
    cn = np.power(10, np.polyval(p.c[:3], neff) + p.c[3]*c)
    gamman = np.polyval(p.gamma[:2], neff) + p.gamma[2]*c
    alphan = np.abs(np.polyval(p.alpha[:3], neff) + p.alpha[3]*c)
    betan = np.polyval(p.beta[:5], neff) + p.beta[5]*c
    mun = np.power(10, np.polyval(p.mu, neff))
    nun = np.power(10, np.polyval(p.nu, neff))

    # Smith et al. 2003 equations C17 & C18
    fa = np.power(omega_m_z, np.asarray(p.fa)[:, np.newaxis, np.newaxis])
    fb = np.power(omega_m_z, np.asarray(p.fb)[:, np.newaxis, np.newaxis])
    f = np.ones((3, np.size(redshift), 1))
    mask = omega_m_z != 1
    fraction = omega_de_z[mask] / (1.0 - omega_m_z[mask])
    f[:, mask] = fraction * fb[:, mask] + (1.0 - fraction) * fa[:, mask]

    # Massive neutrino terms; Bird et al. 2012 equations A6, A9 and A10
    fnu = omega_nu_z / omega_m_z
    Qnu = fnu * (p.l - p.t * (omega_m_z - 0.3)) / (1 + p.m * np.power(y, 3))
    dl2kz = dl2kz * (1 + (p.p * fnu * k2) / (1 + 1.5 * k2))
    betan = betan + fnu * (p.r + p.s * np.square(neff))

    # Two-halo term, Smith et al. 2003 equation C2
    fy = 0.25 * y + 0.125 * np.square(y)
    dq2 = dl2kz * (np.power(1+dl2kz, betan) / (1 + alphan*dl2kz)) * np.exp(-fy)

    # One-halo term, Smith et al. 2003 equations C3 and C4
    # With massive neutrino factor Q_nu, Bird et al. 2012 equation A7
    dh2p = an * np.power(y, 3 * f[0])\
        / (1.0 + bn * np.power(y, f[1]) + np.power(cn * f[2] * y, 3 - gamman))
    dh2 = (1 + Qnu) * dh2p / (1.0 + mun / y + nun / (y * y))

    # Halofit non-linear power spectrum, Smith et al. 2003 equation C1
    pknl = 2 * np.pi * np.pi * (dq2 + dh2) / k3

    return pknl.T.reshape(return_shape)


halofit_smith = partial(halofit, parameters=_smith_parameters)
halofit_takahashi = partial(halofit, parameters=_takahashi_parameters)
halofit_bird = partial(halofit, parameters=_bird_parameters)



"""Abundance matching module.

This module provides methods to perform abundance matching between catalogs of
galaxies and dark matter halos.

Models
======

.. autosummary::
   :nosignatures:
   :toctree: ../api/

   vale_ostriker

"""

from astropy.table import hstack


def vale_ostriker(halos, galaxies, mass='mass', luminosity='luminosity',
                  join_type='inner'):
    """Vale & Ostriker abundance matching.
    Takes catalogs of halos and galaxies and performs abundance matching
    following the method outlined in Vale & Ostriker (2004) assuming
    monotonicity between halo mass and galaxy luminosity to return an
    abundance-matched table of halo-galaxy pairs.

    Parameters
    ----------
    halos, galaxies : Astropy Table
        Tables of halos and galaxies to be matched.
    mass, luminosity : str
        Halo mass and galaxy luminosity column names.
    join_type : str
        Join type (‘inner’ | ‘exact’ | ‘outer’), default is inner

    Returns
    -------
    matched_Table : Astropy Table
        Table of abundance-matched halos and galaxies.

    References
    ----------
    .. [1] Vale A., Ostriker J. P., 2004, MNRAS, 353, 189
    """

    halos.sort(mass, reverse=True)
    galaxies.sort(luminosity, reverse=True)
    return hstack([halos, galaxies], join_type=join_type)



import numpy as np
from astropy.cosmology import Planck15
from astropy.units import allclose
from astropy.utils.data import get_pkg_data_filename
import pytest

# load the external camb result to test against
camb_result_filename = get_pkg_data_filename('data/camb_result.txt')
test_pkz = np.loadtxt(camb_result_filename, delimiter=',')

# try to import the requirement, if it doesn't exist, skip test
try:
    __import__('camb')
except ImportError:
    CAMB_NOT_FOUND = True
else:
    CAMB_NOT_FOUND = False


@pytest.mark.skipif(CAMB_NOT_FOUND, reason='CAMB not found')
def test_camb():
    '''
    Test a default astropy cosmology
    '''
    from skypy.power_spectrum import camb

    redshift = [0.0, 1.0]
    wavenumber = np.logspace(-4.0, np.log10(2.0), 200)
    pkz = camb(wavenumber, redshift, Planck15, 2.e-9, 0.965)
    assert pkz.shape == (len(wavenumber), len(redshift))
    assert allclose(pkz, test_pkz, rtol=1.e-4)

    # also check redshifts are ordered correctly
    redshift = [1.0, 0.0]
    pkz = camb(wavenumber, redshift, Planck15, 2.e-9, 0.965)
    assert pkz.shape == (len(wavenumber), len(redshift))
    assert allclose(pkz, test_pkz[:, np.argsort(redshift)], rtol=1.e-4)



r"""Models of galaxy luminosities.

"""

import numpy as np

import skypy.utils.astronomy as astro
from skypy.utils.random import schechter


__all__ = [
    'herbel_luminosities',
]


def herbel_luminosities(redshift, alpha, a_m, b_m, size=None,
                        x_min=0.00305,
                        x_max=1100.0, resolution=100):

    r"""Model of Herbel et al (2017)

    Luminosities following the Schechter luminosity function following the
    Herbel et al. [1]_ model.

    Parameters
    ----------
    redshift : (nz,) array-like
        The redshift values at which to sample luminosities.
    alpha : float or int
        The alpha parameter in the Schechter luminosity function.
    a_m, b_m : float or int
        Factors parameterising the characteristic absolute magnitude M_* as
        a linear function of redshift according to Equation 3.3 in [1]_.
    size: int, optional
         Output shape of luminosity samples. If size is None and redshift
         is a scalar, a single sample is returned. If size is None and
         redshift is an array, an array of samples is returned with the same
         shape as redshift.
    x_min, x_max : float or int, optional
        Lower and upper luminosity bounds in units of L*.
    resolution : int, optional
        Resolution of the inverse transform sampling spline. Default is 100.

    Returns
    -------
    luminosity : array_like
        Drawn luminosities from the Schechter luminosity function. Luminosities
        are B-band luminosities in units of solar luminosity.

    Notes
    -----
    The Schechter luminosity function is given as

    .. math::
        \Phi(L, z) = \frac{\Phi_\star(z)}{L_\star(z)}
            \left(\frac{L}{L_\star(z)}\right)^\alpha
            \exp\left(-\frac{L}{L_\star(z)}\right) \;.

    Here the luminosity is defined as

    .. math::

        L = 10^{-0.4M} \;,

    with absolute magnitude :math:`M`. Furthermore, Herbel et al. [1]_
    introduced

    .. math::

        \Phi_\star(z) = b_\phi \exp(a_\phi z) \;,
        M_\star(z) = a_M z + b_M \;.

    Now we have to rescale the Schechter function by the comoving element and
    get

    .. math::

        \phi(L,z) = \frac{d_H d_M^2}{E(z)}  \Phi(L,z)\;.

    References
    ----------
    .. [1] Herbel J., Kacprzak T., Amara A. et al., 2017, Journal of Cosmology
        and Astroparticle Physics, Issue 08, article id. 035 (2017)

    Examples
    --------
    >>> import skypy.galaxy.luminosity as lum

    Sample 100 luminosity values at redshift z = 1.0 with
    a_m = -0.9408582, b_m = -20.40492365, alpha = -1.3.

    >>> luminosities = lum.herbel_luminosities(1.0, -1.3, -0.9408582,
    ...                                         -20.40492365, size=100)

    Sample a luminosity value for every redshift in an array z with
    a_m = -0.9408582, b_m = -20.40492365, alpha = -1.3.

    >>> z = np.linspace(0,2, 100)
    >>> luminosities = lum.herbel_luminosities(z, -1.3, -0.9408582,
    ...                                         -20.40492365)

    """

    if size is None and np.shape(redshift):
        size = np.shape(redshift)

    # this is the AB zeropoint of the B-band in units of solar luminosity
    zeropoint = -4.73

    luminosity_star = _calculate_luminosity_star(redshift, a_m, b_m, zeropoint)

    x_sample = schechter(alpha, x_min, x_max, resolution=resolution, size=size)

    return luminosity_star * x_sample


def _calculate_luminosity_star(redshift, a_m, b_m, zeropoint=None):
    absolute_magnitude_star = a_m * redshift + b_m
    return astro.luminosity_from_absolute_magnitude(absolute_magnitude_star, zeropoint)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Skypy is a package offering core functionality and common tools for
astronomical forward-modelling in Python. It contains methods for modelling
the Universe, galaxies and the Milky Way and for generating synthetic
observational data.
"""

# Packages may add whatever they like to this file, but
# should keep this content at the top.
# ----------------------------------------------------------------------------
from ._astropy_init import *   # noqa
# ----------------------------------------------------------------------------

__all__ = []
# from .example_mod import *   # noqa
# Then you can be explicit to control what ends up in the namespace,
#__all__ += ['do_primes']   # noqa
# or you can keep everything from the subpackage with the following instead
# __all__ += example_mod.__all__

from . import cluster  # noqa
from . import galaxy  # noqa
from . import gravitational_wave  # noqa
from . import halo  # noqa
from . import power_spectrum  # noqa
from . import supernova  # noqa



import numpy as np
import scipy.stats
import scipy.integrate
from scipy.special import gammaln
import pytest

from skypy.galaxy import stellar_mass
from skypy.utils import special


def test_exponential_distribution():
    # When alpha=0, M*=1 and x_min~0 we get a truncated exponential
    q_max = 1e2
    sample = stellar_mass.schechter_smf(0, 1, size=1000,
                                        x_min=1e-10, x_max=q_max,
                                        resolution=1000)
    d, p_value = scipy.stats.kstest(sample, 'truncexpon', args=(q_max,))
    assert p_value >= 0.01


def test_stellar_masses():
    # Test that error is returned if m_star input is an array but size !=
    # None and size != m_star,size
    with pytest.raises(ValueError):
        stellar_mass.schechter_smf(-1.4, np.array([1e10, 2e10]), 2, 1, size=3)

    # Test that an array with the sme shape as m_star is returned if m_star is
    # an array and size = None
    m_star = np.array([1e10, 2e10])
    sample = stellar_mass.schechter_smf(-1.4, m_star, 2, 1, size=None)
    assert m_star.shape == sample.shape

    # Test that sampling corresponds to sampling from the right pdf.
    # For this, we sample an array of luminosities for redshift z = 1.0 and we
    # compare it to the corresponding cdf.

    def calc_pdf(m, alpha, mass_star, mass_min, mass_max):
        lg = gammaln(alpha + 1)
        c = np.fabs(special.gammaincc(alpha + 1, mass_min / mass_star))
        d = np.fabs(special.gammaincc(alpha + 1, mass_max / mass_star))
        norm = np.exp(lg) * (c - d)
        return 1. / mass_star * np.power(m / mass_star, alpha) * \
            np.exp(-m / mass_star) / norm

    def calc_cdf(m):
        alpha = -1.4
        mass_star = 10 ** 10.67
        mass_min = 10 ** 7
        mass_max = 10 ** 13
        pdf = calc_pdf(m, alpha, mass_star, mass_min, mass_max)
        cdf = scipy.integrate.cumtrapz(pdf, m, initial=0)
        cdf = cdf / cdf[-1]
        return cdf

    m_star = 10 ** 10.67
    m_min = 10 ** 7
    m_max = 10 ** 13
    sample = stellar_mass.schechter_smf(-1.4, m_star, m_min / m_star,
                                        m_max / m_star, size=1000,
                                        resolution=100)
    p_value = scipy.stats.kstest(sample, calc_cdf)[1]
    assert p_value >= 0.01



import numpy as np
import pytest
from astropy.cosmology import default_cosmology
from skypy.power_spectrum import eisenstein_hu


def test_eisenstein_hu():
    """ Test Eisenstein & Hu Linear matter power spectrum with
    and without wiggles using astropy default cosmology"""
    cosmology = default_cosmology.get()
    A_s = 2.1982e-09
    n_s = 0.969453
    kwmap = 0.02

    # Test that a scalar input gives a scalar output
    scalar_input = 1
    scalar_output_w = eisenstein_hu(scalar_input, A_s, n_s, cosmology, kwmap,
                                    wiggle=True)
    scalar_output_nw = eisenstein_hu(scalar_input, A_s, n_s, cosmology, kwmap,
                                     wiggle=False)
    assert np.isscalar(scalar_output_w)
    assert np.isscalar(scalar_output_nw)

    # Test that an array input gives an array output
    array_shape = (10,)
    array_input = np.random.uniform(size=array_shape)
    array_output_w = eisenstein_hu(array_input, A_s, n_s, cosmology, kwmap,
                                   wiggle=True)
    array_output_nw = eisenstein_hu(array_input, A_s, n_s, cosmology, kwmap,
                                    wiggle=False)
    assert array_output_w.shape == array_shape
    assert array_output_nw.shape == array_shape

    # Test pk against precomputed values for default_cosmology
    wavenumber = np.logspace(-3, 1, num=5, base=10.0)
    pk_eisensteinhu_w = eisenstein_hu(wavenumber, A_s, n_s, cosmology, kwmap,
                                      wiggle=True)
    pk_eisensteinhu_nw = eisenstein_hu(wavenumber, A_s, n_s, cosmology, kwmap,
                                       wiggle=False)
    pk_cosmosis_w = np.array([6.47460158e+03, 3.71610099e+04, 9.65702614e+03,
                              1.14604456e+02, 3.91399918e-01])
    pk_cosmosis_nw = np.array([6.47218600e+03, 3.77330704e+04, 1.00062077e+04,
                              1.13082980e+02, 3.83094714e-01])

    assert np.allclose(pk_eisensteinhu_w, pk_cosmosis_w)
    assert np.allclose(pk_eisensteinhu_nw, pk_cosmosis_nw)

    # Test for failure when wavenumber <= 0
    negative_wavenumber_scalar = 0
    with pytest.raises(ValueError):
        eisenstein_hu(negative_wavenumber_scalar, A_s, n_s, cosmology, kwmap,
                      wiggle=True)
    with pytest.raises(ValueError):
        eisenstein_hu(negative_wavenumber_scalar, A_s, n_s, cosmology, kwmap,
                      wiggle=False)
    negative_wavenumber_array = [0, 1, -2, 3]
    with pytest.raises(ValueError):
        eisenstein_hu(negative_wavenumber_array, A_s, n_s, cosmology, kwmap,
                      wiggle=True)
    with pytest.raises(ValueError):
        eisenstein_hu(negative_wavenumber_array, A_s, n_s, cosmology, kwmap,
                      wiggle=False)



"""Galaxy redshift module.

This module provides facilities to sample galaxy redshifts using a number of
models.
"""

import numpy as np
import scipy.special as sc
import scipy.integrate

import skypy.utils.special as special


__all__ = [
    'herbel_redshift',
    'smail',
]


def smail(z_median, alpha, beta, size=None):
    r'''Redshifts following the Smail et al. (1994) model.

    The redshift follows the Smail et al. [1]_ redshift distribution.

    Parameters
    ----------
    z_median : float or array_like of floats
        Median redshift of the distribution, must be positive.
    alpha : float or array_like of floats
        Power law exponent (z/z0)^\alpha, must be positive.
    beta : float or array_like of floats
        Log-power law exponent exp[-(z/z0)^\beta], must be positive.
    size : None or int or tuple
        Size of the output. If `None`, the size is inferred from the arguments.
        Default is None.

    Notes
    -----
    The probability distribution function :math:`p(z)` for redshift :math:`z`
    is given by Amara & Refregier [2]_ as

    .. math::

        p(z) \sim \left(\frac{z}{z_0}\right)^\alpha
                    \exp\left[-\left(\frac{z}{z_0}\right)^\beta\right] \;.

    This is the generalised gamma distribution.

    References
    ----------
    .. [1] Smail I., Ellis R. S., Fitchett M. J., 1994, MNRAS, 270, 245
    .. [2] Amara A., Refregier A., 2007, MNRAS, 381, 1018

    Examples
    --------
    Sample 10 random variates from the Smail model with `alpha = 1.5` and
    `beta = 2` and median redshift `z_median = 1.2`.

    >>> from skypy.galaxy.redshift import smail
    >>> redshift = smail(1.2, 1.5, 2.0, size=10)

    '''

    k = (alpha+1)/beta
    t = z_median**beta/sc.gammainccinv(k, 0.5)
    g = np.random.gamma(shape=k, scale=t, size=size)
    return g**(1/beta)


def herbel_redshift(alpha, a_phi, b_phi, a_m, b_m, cosmology, low=0.0,
                    high=2.0,
                    size=None, absolute_magnitude_max=-16.0, resolution=100):
    r""" Redshift following the Schechter luminosity function marginalised over
    luminosities following the Herbel et al. [1]_ model.

    Parameters
    ----------
    alpha : float or scalar
        The alpha parameter in the Schechter luminosity function
    a_phi, b_phi : float or scalar
        Parametrisation factors of the normalisation factor Phi_* as a function
        of redshift according to Herbel et al. [1]_ equation (3.4).
    a_m, b_m : float or scalar
        Parametrisation factors of the characteristic absolute magnitude M_* as
        a function of redshift according to Herbel et al. [1]_ equation (3.3).
    cosmology : instance
        Instance of an Astropy Cosmology class.
    low : float or array_like of floats, optional
        The lower boundary of teh output interval. All values generated will be
        greater than or equal to low.
        The default value is 0.0.
    high : float or array_like of floats, optional
        Upper boundary of output interval. All values generated will be less
        than high. The default value is 2.0
    size : int or tuple of ints, optional
        The number of redshifts to sample. If None one value is sampled.
    absolute_magnitude_max : float or scalar, optional
        Upper limit of the considered absolute magnitudes of the galaxies to
        wanna sample.
    resolution : int, optional
        Characterises the resolution of the sampling. Default is 100.

    Returns
    -------
    redshift_sample : ndarray or float
        Drawn redshifts from the marginalised Schechter luminosity function.

    Notes
    -----
    The Schechter luminosity function is given as

    .. math::

        \Phi(L, z) = \frac{\Phi_\star(z)}{L_\star(z)}
            \left(\frac{L}{L_\star(z)}\right)^\alpha
            \exp\left(-\frac{L}{L_\star(z)}\right) \;.

    Here the luminosity is defined as

    .. math::

        L = 10^{-0.4M} \;,

    with absolute magnitude :math:`M`. Furthermore, Herbel et al. [1]_
    introduced

    .. math::

        \Phi_\star(z) = b_\phi \exp(a_\phi z) \;,
        M_\star(z) = a_M z + b_M \;.

    Now we have to rescale the Schechter function by the comoving element and
    get

    .. math::

        \phi(L,z) = \frac{d_H d_M^2}{E(z)}  \Phi(L,z) \;.

     References
    ----------
    .. [1] Herbel J., Kacprzak T., Amara A. et al., 2017, Journal of Cosmology
           and Astroparticle Physics, Issue 08, article id. 035 (2017)

    Examples
    --------
    >>> from skypy.galaxy.redshift import herbel_redshift
    >>> from astropy.cosmology import FlatLambdaCDM

    Sample 100 redshift values from the Schechter luminosity function with
    a_m = -0.9408582, b_m = -20.40492365, a_phi = -0.10268436,
    b_phi = 0.00370253, alpha = -1.3.

    >>> cosmology = FlatLambdaCDM(H0=70, Om0=0.3, Tcmb0=2.725)
    >>> redshift = herbel_redshift(size=1000, low=0.01, alpha=-1.3,
    ...                     a_phi=-0.10268436,a_m=-0.9408582, b_phi=0.00370253,
    ...                     b_m=-20.40492365, cosmology=cosmology,
    ...                     absolute_magnitude_max=-16.)

    """

    redshift = np.linspace(low, high, resolution)
    pdf = herbel_pdf(redshift, alpha, a_phi, b_phi, a_m, b_m, cosmology,
                     absolute_magnitude_max)
    cdf = scipy.integrate.cumtrapz(pdf, redshift, initial=0)
    cdf = cdf / cdf[-1]
    u = np.random.uniform(size=size)
    redshift_sample = np.interp(u, cdf, redshift)

    return redshift_sample


def herbel_pdf(redshift, alpha, a_phi, b_phi, a_m, b_m, cosmology,
               absolute_magnitude_max):
    r"""Calculates the redshift pdf of the Schechter luminosity function
    according to the model of Herbel et al. [1]_ equation (3.6).

    That is, changing the absolute magnitude M in equation (3.2) to luminosity
    L, integrate over all possible L and multiplying by the comoving element
    using a flat :math:`\Lambda \mathrm{CDM}` model to get the corresponding
    pdf.

    Parameters
    ----------
    redshift : array_like
        Input redshifts.
    alpha : float or scalar
        The alpha parameter in the Schechter luminosity function
    a_phi, b_phi : float or scalar
        Parametrisation factors of the normalisation factor Phi_* as a function
        of redshift according to Herbel et al. [1]_ equation (3.4).
    a_m, b_m : float or scalar
        Parametrisation factors of the characteristic absolute magnitude M_* as
        a function of redshift according to Herbel et al. [1]_ equation (3.3).
    cosmology : instance
        Instance of an Astropy Cosmology class.
    luminosity_min : float or scalar
        Cut-off luminosity value such that the Schechter luminosity function
        diverges for L -> 0

    Returns
    -------
    pdf : ndarray or float
    Un-normalised probability density function as a function of redshift
    according to Herbel et al. [1]_.

    Notes
    -----
    This module calculates the function

    .. math::

        \mathrm{pdf}(z) = \Phi_\star(z) \cdot \frac{d_H d_M^2}{E(z)} \cdot
            \Gamma\left(\alpha + 1, \frac{L_\mathrm{min}}{L_\star(z)}\right)\:,

    with :math:`\Phi_\star(z) = b_\phi \exp(a_\phi z)` and the second term the
    comoving element.

    References
    ----------
    .. [1] Herbel J., Kacprzak T., Amara A. et al., 2017, Journal of Cosmology
           and Astroparticle Physics, Issue 08, article id. 035 (2017)

    Examples
    --------
    >>> from skypy.galaxy.redshift import herbel_pdf
    >>> from astropy.cosmology import FlatLambdaCDM
    >>> import numpy as np

    Calculate the pdf for 100 redshift values between 0 and 2 with
    a_m = -0.9408582, b_m = -20.40492365, a_phi = -0.10268436,
    b_phi = 0.00370253, alpha = -1.3 for a flat cosmology.

    >>> cosmology = FlatLambdaCDM(H0=70, Om0=0.3, Tcmb0=2.725)
    >>> redshift = np.linspace(0, 2, 100)
    >>> mag_lim = -16.0
    >>> redshift = herbel_pdf(redshift=redshift, alpha=-1.3,
    ...                     a_phi=-0.10268436,a_m=-0.9408582, b_phi=0.00370253,
    ...                     b_m=-20.40492365, cosmology=cosmology,
    ...                     absolute_magnitude_max=mag_lim)
    """
    abs_mag = a_m*redshift + b_m
    dv = cosmology.differential_comoving_volume(redshift).value
    x = 10.**(-0.4*(absolute_magnitude_max - abs_mag))
    lg = sc.gammaln(alpha+1)
    gx = np.fabs(special.gammaincc(alpha+1, x))
    pdf = dv * b_phi * np.exp(a_phi * redshift + lg) * gx
    return pdf



r"""Galaxy size module.

This modules computes the angular size of galaxies from their physical size.
"""

import numpy as np
from astropy import units


__all__ = [
    'angular_size',
    'early_type_lognormal',
    'late_type_lognormal',
    'linear_lognormal',
]


def angular_size(physical_size, redshift, cosmology):
    """Angular size of a galaxy.
    This function transforms physical radius into angular distance, described
    in [1].

    Parameters
    ----------
    physical_size : astropy.Quantity
        Physical radius of galaxies in units of length.
    redshift : float
        Redshifts at which to evaluate the angular diameter distance.
    cosmology : astropy.cosmology.Cosmology
        Cosmology object providing methods for the evolution history of
        omega_matter and omega_lambda with redshift.

    Returns
    -------
    angular_size : astropy.Quantity
        Angular distances in units of [rad] for a given radius.

    Examples
    --------
    >>> from astropy import units
    >>> from astropy.cosmology import FlatLambdaCDM
    >>> cosmology = FlatLambdaCDM(H0=67.04, Om0=0.3183, Ob0=0.047745)
    >>> size = 10.0 * units.kpc
    >>> angular_size(size, 1, cosmology)
    <Quantity 5.85990062e-06 rad>

    References
    ----------
    .. [1] D. W. Hogg, (1999), astro-ph/9905116.
    """

    distance = cosmology.angular_diameter_distance(redshift)
    angular_size = np.arctan(physical_size / distance)

    return angular_size


def late_type_lognormal(magnitude, alpha, beta, gamma, M0, sigma1, sigma2,
                        size=None):
    """Lognormal distribution for late-type galaxies.
    This function provides a lognormal distribution for the physical size of
    late-type galaxies, described by equations 12, 15 and 16 in [1].

    Parameters
    ----------
    magnitude : float or array_like.
        Galaxy magnitude at which evaluate the lognormal distribution.
    alpha, beta, gamma, M0: float
        Model parameters describing the mean size of galaxies in [kpc].
        Equation 15 in [1].
    sigma1, sigma2: float
        Parameters describing the standard deviation of the lognormal
        distribution for the physical radius of galaxies. Equation 16 in [1].
    size : int or tuple of ints, optional.
        Output shape. If the given shape is, e.g., (m, n, k),
        then m * n * k samples are drawn. If size is None (default),
        a single value is returned if mean and sigma are both scalars.
        Otherwise, np.broadcast(mean, sigma).size samples are drawn.

    Returns
    -------
    physical_size : numpy.ndarray or astropy.Quantity
        Physical distance for a given galaxy with a given magnitude, in [kpc].
        If size is None and magnitude is a scalar, a single sample is returned.
        If size is ns, different from None, and magnitude is scalar,
        shape is (ns,). If magnitude has shape (nm,) and size=None,
        shape is (nm,).

    Examples
    --------
    >>> import numpy as np
    >>> from skypy.galaxy import size
    >>> np.random.seed(12345)
    >>> magnitude = -16.0
    >>> alpha, beta, gamma, M0 = 0.21, 0.53, -1.31, -20.52
    >>> sigma1, sigma2 = 0.48, 0.25
    >>> size.late_type_lognormal(magnitude, alpha, beta, gamma, M0,\
                                 sigma1, sigma2)
    <Quantity 0.9850926 kpc>


    References
    ----------
    ..[1] S. Shen, H.J. Mo, S.D.M. White, M.R. Blanton, G. Kauffmann, W. Voges,
        J. Brinkmann, I. Csabai, Mon. Not. Roy. Astron. Soc. 343, 978 (2003).
    """

    if size is None and np.shape(magnitude):
        size = np.shape(magnitude)

    r_bar = np.power(10, -0.4 * alpha * magnitude + (beta - alpha) *
                     np.log10(1 + np.power(10, -0.4 * (magnitude - M0)))
                     + gamma) * units.kpc

    sigma_lnR = sigma2 + (sigma1 - sigma2) /\
                         (1.0 + np.power(10, -0.8 * (magnitude - M0)))

    return r_bar * np.random.lognormal(sigma=sigma_lnR, size=size)


def early_type_lognormal(magnitude, a, b, M0, sigma1, sigma2, size=None):
    """Lognormal distribution for early-type galaxies.
    This function provides a lognormal distribution for the physical size of
    early-type galaxies, described by equations 12, 14 and 16 in [1].

    Parameters
    ----------
    magnitude : float or array_like.
        Galaxy magnitude at which evaluate the lognormal distribution.
    a, b : float
        Linear model parameters describing the mean size of galaxies,
        equation 3.14 in [1].
    sigma: float
        Standard deviation of the lognormal distribution for the
        physical radius of galaxies.
    size : int or tuple of ints, optional.
        Output shape. If the given shape is, e.g., (m, n, k),
        then m * n * k samples are drawn. If size is None (default),
        a single value is returned if mean and sigma are both scalars.
        Otherwise, np.broadcast(mean, sigma).size samples are drawn.

    Returns
    -------
    physical_size : ndarray or astropy.Quantity
        Physical distance for a given galaxy with a given magnitude, in [kpc].
        If size is None and magnitude is a scalar, a single sample is returned.
        If size is ns, different from None, and magnitude is scalar,
        shape is (ns,). If magnitude has shape (nm,) and size=None,
        shape is (nm,).

    Examples
    --------
    >>> import numpy as np
    >>> from skypy.galaxy import size
    >>> np.random.seed(12345)
    >>> magnitude = -20.0
    >>> a, b, M0 = 0.6, -4.63, -20.52
    >>> sigma1, sigma2 = 0.48, 0.25
    >>> size.early_type_lognormal(magnitude, a, b, M0, sigma1, sigma2)
    <Quantity 1.35830285 kpc>


    References
    ----------
    ..[1] S. Shen, H.J. Mo, S.D.M. White, M.R. Blanton, G. Kauffmann, W. Voges,
        J. Brinkmann, I. Csabai, Mon. Not. Roy. Astron. Soc. 343, 978 (2003).
    """

    return late_type_lognormal(magnitude, a, a, b, M0, sigma1, sigma2,
                               size=size)


def linear_lognormal(magnitude, a_mu, b_mu, sigma, size=None):
    """Lognormal distribution with linear mean.
    This function provides a lognormal distribution for the physical size of
    galaxies with a linear mean, described by equation 3.14 in [1]. See also
    equation 14 in [2].

    Parameters
    ----------
    magnitude : float or array_like.
        Galaxy absolute magnitude at which evaluate the lognormal distribution.
    a_mu, b_mu : float
        Linear model parameters describing the mean size of galaxies,
        equation 3.14 in [1].
    sigma: float
        Standard deviation of the lognormal distribution for the
        physical radius of galaxies.
    size : int or tuple of ints, optional.
        Output shape. If the given shape is, e.g., (m, n, k),
        then m * n * k samples are drawn. If size is None (default),
        a single value is returned if mean and sigma are both scalars.
        Otherwise, np.broadcast(mean, sigma).size samples are drawn.

    Returns
    -------
    physical_size : numpy.ndarray or astropy.Quantity
        Physical distance for a given galaxy with a given magnitude, in [kpc].
        If size is None and magnitude is a scalar, a single sample is returned.
        If size is ns, different from None, and magnitude is scalar,
        shape is (ns,). If magnitude has shape (nm,) and size=None,
        shape is (nm,).

    Examples
    --------
    >>> import numpy as np
    >>> from skypy.galaxy import size
    >>> np.random.seed(12345)
    >>> magnitude = -20.0
    >>> a_mu, b_mu, sigma =-0.24, -4.63, 0.4
    >>> size.linear_lognormal(magnitude, a_mu, b_mu, sigma)
    <Quantity 1.36282044 kpc>

    References
    ----------
    ..[1] J. Herbel, T. Kacprzak, A. Amara, A. Refregier, C.Bruderer and
           A. Nicola, JCAP 1708, 035 (2017).
    ..[2] S. Shen, H.J. Mo, S.D.M. White, M.R. Blanton, G. Kauffmann, W.Voges,
           J. Brinkmann, I.Csabai, Mon. Not. Roy. Astron. Soc. 343, 978 (2003).
    """

    return late_type_lognormal(magnitude, -a_mu / 0.4, -a_mu / 0.4,
                               b_mu, -np.inf, sigma, sigma, size=size)



"""
This module contains methods that model the intrinsic properties of galaxy
populations.
"""

from . import ellipticity  # noqa F401,F403
from . import luminosity  # noqa F401,F403
from . import redshift  # noqa F401,F403
from . import size  # noqa F401,F403
from . import spectrum  # noqa F401,F403
from . import stellar_mass  # noqa F401,F403



import numpy as np
from astropy.cosmology import Planck15
from astropy.units import allclose
from skypy.power_spectrum import eisenstein_hu


import skypy.halo.mass as mass

# Precomputed values for the test, for a Planck15 cosmology at redshift 0 and a
# power spectrum given by the Eisenstein and Hu fitting formula
# Models: Press-Schechter and Sheth-Tormen
mass_array = [1.00000000e+10, 3.16227766e+10, 1.00000000e+11, 3.16227766e+11,
              1.00000000e+12, 3.16227766e+12, 1.00000000e+13, 3.16227766e+13,
              1.00000000e+14, 3.16227766e+14]
ST_fsigma = [0.22851856, 0.24659885, 0.27788, 0.30138876, 0.31338251,
             0.32004239, 0.31068427, 0.2676586, 0.16714694, 0.0957242]
PS_fsigma = [0.27373595, 0.31173523, 0.38096078, 0.4358908, 0.4651783,
             0.48375949, 0.46491057, 0.37505525, 0.18939675, 0.08453155]
E_fsigma = [0.21224081, 0.22905786, 0.25820042, 0.28018667, 0.29148879,
            0.29805329, 0.28973637, 0.25038455, 0.1574158, 0.09077113]
ST_massf = [5.91918389e-12, 2.04006848e-12, 3.90033154e-13, 3.43105199e-14,
            2.47935792e-15, 3.85619829e-16, 2.77447365e-17, 3.53176402e-18,
            2.69047690e-19, 1.13766655e-20]
PS_massf = [7.09042387e-12, 2.57893025e-12, 5.34717624e-13, 4.96224201e-14,
            3.68030594e-15, 5.82882944e-16, 4.15174583e-17, 4.94886632e-18,
            3.04862034e-19, 1.00464376e-20]
E_massf = [5.49755090e-12, 1.89495500e-12, 3.62410841e-13, 3.18968432e-14,
           2.30614353e-15, 3.59125102e-16, 2.58740463e-17, 3.30383237e-18,
           2.53383973e-19, 1.07880011e-20]

cosmo = Planck15
k = np.logspace(-3, 1, num=10, base=10.0)
A_s, n_s = 2.1982e-09, 0.969453
Pk = eisenstein_hu(k, A_s, n_s, cosmo, kwmap=0.02, wiggle=True)


def test_halo_mass_function():
    # Test the output and shape is correct given an array of masses
    m_array = np.asarray(mass_array)

    # Any particular ellipsoidal collapse model
    params_E = (0.3, 0.7, 0.3, 1.686)
    fE = mass.ellipsoidal_collapse_function
    array_output_E = mass.halo_mass_function(m_array, k, Pk, 1.0, cosmo,
                                             fE, params=params_E)
    assert array_output_E.shape == m_array.shape
    assert allclose(array_output_E, E_massf)

    # Sheth and Tormen collapse model
    array_output_ST = mass.sheth_tormen_mass_function(m_array, k, Pk,
                                                      1.0, cosmo)
    assert array_output_ST.shape == m_array.shape
    assert allclose(array_output_ST, ST_massf)

    # Press-Schechter collapse model
    array_output_PS = mass.press_schechter_mass_function(m_array, k, Pk,
                                                         1.0, cosmo)
    assert array_output_PS.shape == m_array.shape
    assert allclose(array_output_PS, PS_massf)


def test_halo_mass_sampler():
    # Test the output shape is correct given the sample size
    n_samples = 1000
    m_min, m_max, resolution = 10**9, 10**12, 100
    # Any particular ellipsoidal collapse model
    params_E = (0.3, 0.7, 0.3, 1.686)
    fE = mass.ellipsoidal_collapse_function
    array_output_E = mass.halo_mass_sampler(m_min, m_max, resolution, k, Pk,
                                            1.0, cosmo, fE, params=params_E,
                                            size=n_samples)
    assert len(array_output_E) == n_samples

    # Sheth and Tormen collapse model
    array_output_PS = mass.sheth_tormen(10**9, 10**12, 100, k,
                                        Pk, 1.0, cosmo, size=n_samples)

    assert len(array_output_PS) == n_samples

    # Press-Schechter collapse model
    array_output_PS = mass.press_schechter(10**9, 10**12, 100, k,
                                           Pk, 1.0, cosmo, size=n_samples)

    assert len(array_output_PS) == n_samples


def test_ellipsoidal_collapse_function():
    # Test any ellipsoidal model against precomputed values
    m_array = np.asarray(mass_array)
    sigma = np.sqrt(mass._sigma_squared(m_array, k, Pk, 1.0, cosmo))
    params_E = (0.3, 0.7, 0.3, 1.686)
    fE = mass.ellipsoidal_collapse_function(sigma, params_E)
    assert allclose(fE, E_fsigma)

    # Test the Sheth and Tormen model against precomputed values
    m_array = np.asarray(mass_array)
    sigma = np.sqrt(mass._sigma_squared(m_array, k, Pk, 1.0, cosmo))
    fst = mass.sheth_tormen_collapse_function(sigma)
    assert allclose(fst, ST_fsigma)

    # Test the Press-Schechter model against precomputed values
    m_array = np.asarray(mass_array)
    sigma = np.sqrt(mass._sigma_squared(m_array, k, Pk, 1.0, cosmo))
    fps = mass.press_schechter_collapse_function(sigma)
    assert allclose(fps, PS_fsigma)



from astropy.table import Table
from astropy.table.np_utils import TableMergeError
import numpy as np
import pytest
from skypy.halo.abundance_matching import vale_ostriker


def test_vale_ostriker():
    """ Test Vale & Ostriker abundance matching algorithm"""

    # Mock catalogs for testing
    nh = 20
    ng = 10
    h = Table(data=np.random.uniform(size=(nh, 2)), names=['M', 'mass'])
    g = Table(data=np.random.uniform(size=(ng, 2)), names=['L', 'luminosity'])

    # Test default keys
    matched_default = vale_ostriker(h, g)
    argsort_m = np.argsort(matched_default['mass'])
    argsort_l = np.argsort(matched_default['luminosity'])
    assert np.all(argsort_m == argsort_l)

    # Test custom keys
    matched_custom = vale_ostriker(h, g, mass='M', luminosity='L')
    argsort_m = np.argsort(matched_custom['M'])
    argsort_l = np.argsort(matched_custom['L'])
    assert np.all(argsort_m == argsort_l)

    # Test failure with bad keys
    with pytest.raises(ValueError):
        vale_ostriker(h, g, mass='bad_key')
    with pytest.raises(ValueError):
        vale_ostriker(h, g, luminosity='bad_key')

    # Test correct table lengths for inner and outer joins
    assert len(vale_ostriker(h, g, join_type='inner')) == min(nh, ng)
    assert len(vale_ostriker(h, g, join_type='outer')) == max(nh, ng)

    # Test failure for exact join when nh != ng
    with pytest.raises(TableMergeError):
        vale_ostriker(h, g, join_type='exact')



import numpy as np
from astropy import units as u


__all__ = [
    'camb',
]


def camb(wavenumber, redshift, cosmology, A_s, n_s):
    """ Return the CAMB computation of the linear matter power spectrum, on a
    two dimensional grid of wavenumber and redshift

    Parameters
    ----------
    wavenumber : (nk,) array_like
        Array of wavenumbers in units of [Mpc^-1] at which to
        evaluate the linear matter power spectrum.
    redshift : (nz,) array_like
        Array of redshifts at which to evaluate the linear matter power
        spectrum.
    cosmology : astropy.cosmology.Cosmology
        Cosmology object providing omega_matter, omega_baryon, Hubble
        parameter and CMB temperature in the present day
    A_s : float
        Cosmology parameter, amplitude normalisation of curvature perturbation
        power spectrum
    n_s : float
        Cosmology parameter, spectral index of scalar perturbation power
        spectrum

    Returns
    -------
    power_spectrum : (nk, nz) array_like
        Array of values for the linear matter power spectrum in  [Mpc^3]
        evaluated at the input wavenumbers for the given primordial power
        spectrum parameters, cosmology. For nz redshifts and nk wavenumbers
        the returned array will have shape (nk, nz).

    Examples
    --------
    >>> import numpy as np
    >>> from astropy.cosmology import default_cosmology
    >>> cosmology = default_cosmology.get()
    >>> redshift = np.array([0, 1])
    >>> wavenumber = np.array([1.e-2, 1.e-1, 1e0])
    >>> A_s = 2.e-9
    >>> n_s = 0.965
    >>> camb(wavenumber, redshift, cosmology, A_s, n_s)  # doctest: +SKIP
    array([[2.34758952e+04, 8.70837957e+03],
           [3.03660813e+03, 1.12836115e+03],
           [2.53124880e+01, 9.40802814e+00]])

    References
    ----------
    doi : 10.1086/309179
    arXiv: astro-ph/9911177

    """

    try:
        from camb import CAMBparams, get_results, model
    except ImportError:
        raise Exception("camb is required to use skypy.linear.camb")

    redshift = np.atleast_1d(redshift)

    h2 = cosmology.h * cosmology.h

    pars = CAMBparams()
    pars.set_cosmology(H0=cosmology.H0.value,
                       ombh2=cosmology.Ob0 * h2,
                       omch2=cosmology.Odm0 * h2,
                       omk=cosmology.Ok0,
                       TCMB=cosmology.Tcmb0.value,
                       mnu=np.sum(cosmology.m_nu.value),
                       standard_neutrino_neff=cosmology.Neff
                       )

    # camb requires redshifts to be in decreasing order
    redshift_order = np.argsort(redshift)[::-1]

    pars.InitPower.ns = n_s
    pars.InitPower.As = A_s

    pars.set_matter_power(redshifts=list(redshift[redshift_order]),
                          kmax=np.max(wavenumber))

    pars.NonLinear = model.NonLinear_none

    results = get_results(pars)

    k = wavenumber * (1. / u.Mpc)

    k_h = k.to((u.littleh / u.Mpc), u.with_H0(cosmology.H0))

    kh, z, pzk = results.get_matter_power_spectrum(minkh=np.min(k_h.value),
                                                   maxkh=np.max(k_h.value),
                                                   npoints=len(k_h.value))

    return pzk[redshift_order[::-1]].T



import numpy as np
from astropy.cosmology import FlatLambdaCDM, default_cosmology, Planck15
from astropy.units import isclose, allclose
import pytest

from skypy.power_spectrum import (growth_function_carroll, growth_factor,
                                  growth_function, growth_function_derivative)


def test_carroll():
    """ Test a FlatLambdaCDM cosmology with omega_matter = 1"""
    cosmology = FlatLambdaCDM(Om0=1.0, H0=70.0)

    # Test that a scalar input gives a scalar output
    scalar_input = 1
    scalar_output = growth_function_carroll(scalar_input, cosmology)
    assert np.isscalar(scalar_output)

    # Test that an array input gives an array output
    array_shape = (10,)
    array_input = np.random.uniform(size=array_shape)
    array_output = growth_function_carroll(array_input, cosmology)
    assert array_output.shape == array_shape

    # Test against theory for omega_matter = 1.0
    redshift = np.linspace(0.0, 10.0, 100)
    Dz_carroll = growth_function_carroll(redshift, cosmology)
    Dz_theory = 1.0 / (1.0 + redshift)
    assert allclose(Dz_carroll, Dz_theory)

    # Test against precomputed values for Planck15
    redshift = np.linspace(0, 5, 4)
    Dz_carroll = growth_function_carroll(redshift, Planck15)
    Dz_truth = np.array([0.78136173, 0.36635322, 0.22889793, 0.16577711])
    assert allclose(Dz_carroll, Dz_truth)

    # Test for failure when redshift < 0
    negative_redshift_scalar = -1
    with pytest.raises(ValueError):
        growth_function_carroll(negative_redshift_scalar, cosmology)
    negative_redshift_array = [0, 1, -2, 3]
    with pytest.raises(ValueError):
        growth_function_carroll(negative_redshift_array, cosmology)


def test_growth():
    """
    Test a FlatLambdaCDM cosmology with omega_matter = 1.0
    and astropy default cosmology
    """

    redshift = np.linspace(0., 10., 101)
    cosmology_flat = FlatLambdaCDM(H0=70.0, Om0=1.0)

    fz = growth_factor(redshift, cosmology_flat)
    Dz = growth_function(redshift, cosmology_flat)
    Dzprime = growth_function_derivative(redshift, cosmology_flat)

    # Test growth factor
    assert redshift.shape == fz.shape,\
        "Length of redshift array and growth rate array do not match"
    assert isclose(fz[0], 1.0),\
        "Growth factor at redshift 0 is not close to 1.0"

    # Test growth function
    assert redshift.shape == Dz.shape,\
        "Length of redshift array and growth function array do not match"
    assert allclose(Dz, 1. / (1. + redshift)),\
        "Growth function is not close to the scale factor"

    # Test growth function derivative
    assert redshift.shape == Dzprime.shape,\
        "Length of redshift array and growth function array do not match"
    assert isclose(Dzprime[0], -1.0),\
        "Derivative of growth function at redshift 0 is not close to -1.0"

    # Test against precomputed values using astropy default cosmology
    default = default_cosmology.get()
    zvec = np.linspace(0.0, 1.0, 4)

    fz_default = growth_factor(zvec, default)
    Dz_default = growth_function(zvec, default)
    Dzprime_default = growth_function_derivative(zvec, default)

    precomputed_fz_default = np.array([0.5255848, 0.69412802, 0.80439553,
                                       0.87179376])
    precomputed_Dz_default = np.array([0.66328939, 0.55638978, 0.4704842,
                                       0.40368459])
    precomputed_Dzprime_default = np.array([-0.34861482, -0.2896543,
                                            -0.22707323, -0.17596485])

    assert allclose(fz_default, precomputed_fz_default)
    assert allclose(Dz_default, precomputed_Dz_default)
    assert allclose(Dzprime_default, precomputed_Dzprime_default)



"""
This module contains methods that model the matter power spectrum.

Linear Power Spectrum
=====================

.. autosummary::
   :nosignatures:
   :toctree: ../api/

   camb
   eisenstein_hu
   transfer_no_wiggles
   transfer_with_wiggles


Nonlinear Power Spectrum
========================

.. autosummary::
   :nosignatures:
   :toctree: ../api/

   HalofitParameters
   halofit
   halofit_smith
   halofit_takahashi
   halofit_bird


Growth Functions
================

.. autosummary::
   :nosignatures:
   :toctree: ../api/

   growth_factor
   growth_function
   growth_function_carroll
   growth_function_derivative

"""

from ._camb import *  # noqa F401,F403
from ._eisenstein_hu import *  # noqa F401,F403
from ._halofit import *  # noqa F401,F403
from ._growth import *  # noqa F401,F403



import numpy as np
from scipy import stats
from collections import Counter

from skypy.halo.quenching import environment_quenched, mass_quenched


def test_environment_quenched():
    # Test the quenching process follows a binomial distribution
    n, p = 1000, 0.7
    quenched = environment_quenched(n, p)
    number_quenched = Counter(quenched)[True]

    p_value = stats.binom_test(number_quenched, n=n, p=p,
                               alternative='two-sided')
    assert p_value > 0.05


def test_mass_quenched():
    # Test the quenching process follows a binomial distribution if the
    # logarithmic halo masses are symmetrical about the offset
    n = 1000
    offset, width = 1.0e12, 0.5
    halo_mass = 10 ** np.random.uniform(11, 13, n)
    quenched = mass_quenched(halo_mass, offset, width)
    number_quenched = Counter(quenched)[True]

    p_value = stats.binom_test(number_quenched, n=n, p=0.5,
                               alternative='two-sided')
    assert p_value > 0.05
